/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import java.nio.charset.Charset;
import java.nio.charset.UnsupportedCharsetException;

import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.beacon.BeaconDetectorConfiguration;
import com.aquenos.epics.jackie.client.resolver.ChannelNameResolver;
import com.aquenos.epics.jackie.client.resolver.ChannelNameResolverConfiguration;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.exception.ExceptionSafeErrorHandler;
import com.aquenos.epics.jackie.common.exception.JavaUtilLoggingErrorHandler;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessageCodec;
import com.aquenos.epics.jackie.common.util.EnvironmentUtil;
import com.aquenos.epics.jackie.common.util.HostNameUtil;
import com.aquenos.epics.jackie.common.util.UserNameUtil;

/**
 * Configuration for a {@link ChannelAccessClient}.
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessClientConfiguration {

    private Charset charset;
    private double echoInterval;
    private String hostName;
    private String userName;
    private int maxPayloadReceiveSize;
    private int maxPayloadSendSize;
    private ChannelAccessEventMask defaultEventMask;
    private long cidBlockReuseTime;
    private ErrorHandler errorHandler;
    private BeaconDetectorConfiguration beaconDetectorConfiguration;
    private ChannelNameResolverConfiguration resolverConfiguration;
    private ErrorHandler wrappedErrorHandler;

    /**
     * Creates a configuration for a Channel Access client that uses default
     * values for all options. This is the same as passing <code>null</code> for
     * all parameters of the
     * {@link #ChannelAccessClientConfiguration(Charset, String, String, Integer, Integer, Double, Long, ChannelAccessEventMask, ErrorHandler, BeaconDetectorConfiguration, ChannelNameResolverConfiguration)}
     * constructor.
     */
    public ChannelAccessClientConfiguration() {
        this(null, null, null, null, null, null, null, null, null, null, null);
    }

    /**
     * Creates a configuration for a Channel Access client using the specified
     * options. If <code>null</code> is passed for an option, the default value
     * for this option is used.
     * 
     * @param charset
     *            character set used to encode and decode strings. If this
     *            parameter is <code>null</code> the UTF-8 character set is
     *            used. If the UTF-8 character set is not available on this
     *            platform, the platform's default character set is used.
     * @param hostName
     *            host name identifying the client computer when connecting to a
     *            server. If <code>null</code> the host name is determined
     *            automatically using the {@link HostNameUtil#getHostName()}
     *            method. If the host name cannot be determined automatically,
     *            the special string "&lt;unknown host&gt;" is used.
     * @param userName
     *            user name identifying the user when connecting to a server. If
     *            <code>null</code> the user name is determined automatically
     *            using the {@link UserNameUtil#getUserName()} method. If the
     *            user name cannot be determined automatically, the empty string
     *            is used.
     * @param maxPayloadReceiveSize
     *            maximum size of the payload of a message received via TCP (in
     *            bytes). This effectively limits the number of elements that
     *            can be transferred from the server in a get operation for a
     *            channel. Must be at least 16384 bytes. If <code>null</code> a
     *            default value of 2147483616 is used, unless the
     *            <code>EPICS_CA_AUTO_ARRAY_BYTES</code> environment variable is
     *            set to <code>NO</code>. In that case, the
     *            <code>EPICS_CA_MAX_ARRAY_BYTES</code> environment variable is
     *            used and if it is not set or its value is invalid, a default
     *            value of 16384 bytes is used.
     * @param maxPayloadSendSize
     *            maximum size of the payload of a message sent via TCP (in
     *            bytes). This effectively limits the number of elements that
     *            can be transferred to the server in a put operation for a
     *            channel. Must be at least 16384 bytes. If <code>null</code> a
     *            default value of 2147483616 is used, unless the
     *            <code>EPICS_CA_AUTO_ARRAY_BYTES</code> environment variable is
     *            set to <code>NO</code>. In that case, the
     *            <code>EPICS_CA_MAX_ARRAY_BYTES</code> environment variable is
     *            used and if it is not set or its value is invalid, a default
     *            value of 16384 bytes is used.
     * @param cidBlockReuseTime
     *            time that a CID is blocked from being used again in
     *            milliseconds. If <code>null</code> a default value of 900000
     *            milliseconds (15 minutes) is used. After destroying a channel,
     *            the CID may not be reused for some time because there might
     *            still be late responses to old search requests, which would be
     *            used for the wrong channel if the CID was reused too early. A
     *            value of 0 (or a negative value) means that CIDs can be reused
     *            immediately.
     * @param defaultEventMask
     *            default event mask to use when creating monitors. If
     *            <code>null</code> an event mask of
     *            <code>DBE_VALUE|DBE_ALARM</code> is used. May not be
     *            <code>DBE_NONE</code>.
     * @param errorHandler
     *            error handler to use for handling all errors in the client or
     *            <code>null</code> if the default error handler (
     *            {@link JavaUtilLoggingErrorHandler}) should be used.
     * @param beaconDetectorConfiguration
     *            configuration for the {@link BeaconDetector} that should be
     *            used or <code>null</code> to use a configuration with default
     *            settings. If this parameter is <code>null</code>, the
     *            configuration that is created will use the specified
     *            <code>errorHandler</code>.
     * @param resolverConfiguration
     *            configuration for the {@link ChannelNameResolver} that should
     *            be used or <code>null</code> to use a configuration with
     *            default settings. If this parameter is <code>null</code>, the
     *            configuration that is created will use the specified
     *            <code>charset</code> and <code>errorHandler</code>.
     * @throws IllegalArgumentException
     *             if <code>maxPayloadReceiveSize</code> or
     *             <code>maxPayloadSendSize</code> are less than 16384 or
     *             greater than 2147483616 or <code>defaultEventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @deprecated Since 2.0.0. Use
     *             {@link #ChannelAccessClientConfiguration(Charset, String, String, Integer, Integer, Double, Long, ChannelAccessEventMask, ErrorHandler, BeaconDetectorConfiguration, ChannelNameResolverConfiguration)}
     *             instead.
     */
    public ChannelAccessClientConfiguration(Charset charset, String hostName,
            String userName, Integer maxPayloadReceiveSize,
            Integer maxPayloadSendSize, Long cidBlockReuseTime,
            ChannelAccessEventMask defaultEventMask, ErrorHandler errorHandler,
            BeaconDetectorConfiguration beaconDetectorConfiguration,
            ChannelNameResolverConfiguration resolverConfiguration) {
        this(charset, hostName, userName, maxPayloadReceiveSize,
                maxPayloadSendSize, null, cidBlockReuseTime, defaultEventMask,
                errorHandler, beaconDetectorConfiguration,
                resolverConfiguration);
    }

    /**
     * Creates a configuration for a Channel Access client using the specified
     * options. If <code>null</code> is passed for an option, the default value
     * for this option is used.
     * 
     * @param charset
     *            character set used to encode and decode strings. If this
     *            parameter is <code>null</code> the UTF-8 character set is
     *            used. If the UTF-8 character set is not available on this
     *            platform, the platform's default character set is used.
     * @param hostName
     *            host name identifying the client computer when connecting to a
     *            server. If <code>null</code> the host name is determined
     *            automatically using the {@link HostNameUtil#getHostName()}
     *            method. If the host name cannot be determined automatically,
     *            the special string "&lt;unknown host&gt;" is used.
     * @param userName
     *            user name identifying the user when connecting to a server. If
     *            <code>null</code> the user name is determined automatically
     *            using the {@link UserNameUtil#getUserName()} method. If the
     *            user name cannot be determined automatically, the empty string
     *            is used.
     * @param maxPayloadReceiveSize
     *            maximum size of the payload of a message received via TCP (in
     *            bytes). This effectively limits the number of elements that
     *            can be transferred from the server in a get operation for a
     *            channel. Must be at least 16384 bytes. If <code>null</code> a
     *            default value of 2147483616 is used, unless the
     *            <code>EPICS_CA_AUTO_ARRAY_BYTES</code> environment variable is
     *            set to <code>NO</code>. In that case, the
     *            <code>EPICS_CA_MAX_ARRAY_BYTES</code> environment variable is
     *            used and if it is not set or its value is invalid, a default
     *            value of 16384 bytes is used.
     * @param maxPayloadSendSize
     *            maximum size of the payload of a message sent via TCP (in
     *            bytes). This effectively limits the number of elements that
     *            can be transferred to the server in a put operation for a
     *            channel. Must be at least 16384 bytes. If <code>null</code> a
     *            default value of 2147483616 is used, unless the
     *            <code>EPICS_CA_AUTO_ARRAY_BYTES</code> environment variable is
     *            set to <code>NO</code>. In that case, the
     *            <code>EPICS_CA_MAX_ARRAY_BYTES</code> environment variable is
     *            used and if it is not set or its value is invalid, a default
     *            value of 16384 bytes is used.
     * @param echoInterval
     *            interval between sending echo requests to a server in seconds.
     *            If <code>null</code>, the<code>EPICS_CA_CONN_TMO</code>
     *            environment variable is used. If the environment variable is
     *            not set or its value is invalid, a default value of 30 seconds
     *            is used.
     * @param cidBlockReuseTime
     *            time that a CID is blocked from being used again in
     *            milliseconds. If <code>null</code> a default value of 900000
     *            milliseconds (15 minutes) is used. After destroying a channel,
     *            the CID may not be reused for some time because there might
     *            still be late responses to old search requests, which would be
     *            used for the wrong channel if the CID was reused too early. A
     *            value of 0 (or a negative value) means that CIDs can be reused
     *            immediately.
     * @param defaultEventMask
     *            default event mask to use when creating monitors. If
     *            <code>null</code> an event mask of
     *            <code>DBE_VALUE|DBE_ALARM</code> is used. May not be
     *            <code>DBE_NONE</code>.
     * @param errorHandler
     *            error handler to use for handling all errors in the client or
     *            <code>null</code> if the default error handler (
     *            {@link JavaUtilLoggingErrorHandler}) should be used.
     * @param beaconDetectorConfiguration
     *            configuration for the {@link BeaconDetector} that should be
     *            used or <code>null</code> to use a configuration with default
     *            settings. If this parameter is <code>null</code>, the
     *            configuration that is created will use the specified
     *            <code>errorHandler</code>.
     * @param resolverConfiguration
     *            configuration for the {@link ChannelNameResolver} that should
     *            be used or <code>null</code> to use a configuration with
     *            default settings. If this parameter is <code>null</code>, the
     *            configuration that is created will use the specified
     *            <code>charset</code>, <code>errorHandler</code>. and
     *            <code>echoInterval</code>.
     * @throws IllegalArgumentException
     *             if <code>maxPayloadReceiveSize</code> or
     *             <code>maxPayloadSendSize</code> are less than 16384,
     *             <code>echoInterval</code> is less than 100, or
     *             <code>defaultEventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @since EPICS Jackie 2.0.0
     */
    public ChannelAccessClientConfiguration(Charset charset, String hostName,
            String userName, Integer maxPayloadReceiveSize,
            Integer maxPayloadSendSize, Double echoInterval,
            Long cidBlockReuseTime, ChannelAccessEventMask defaultEventMask,
            ErrorHandler errorHandler,
            BeaconDetectorConfiguration beaconDetectorConfiguration,
            ChannelNameResolverConfiguration resolverConfiguration) {
        if (charset != null) {
            this.charset = charset;
        } else {
            try {
                this.charset = Charset.forName("UTF-8");
            } catch (UnsupportedCharsetException e) {
                this.charset = Charset.defaultCharset();
            }
        }
        if (hostName != null) {
            this.hostName = hostName;
        } else {
            initializeHostNameFromEnvironment();
        }
        if (userName != null) {
            this.userName = userName;
        } else {
            initializeUserNameFromEnvironment();
        }
        if (maxPayloadReceiveSize != null) {
            if (maxPayloadReceiveSize < 16384
                    || maxPayloadReceiveSize > ChannelAccessMessageCodec.MAX_MAX_PAYLOAD_SIZE) {
                throw new IllegalArgumentException(
                        "The maximum payload send size must be at least 16384 bytes and at most "
                                + ChannelAccessMessageCodec.MAX_MAX_PAYLOAD_SIZE
                                + ".");
            }
            this.maxPayloadReceiveSize = maxPayloadReceiveSize;
        } else {
            initializeMaxPayloadReceiveSizeFromEnvironment();
        }
        if (maxPayloadSendSize != null) {
            if (maxPayloadSendSize < 16384
                    || maxPayloadSendSize > ChannelAccessMessageCodec.MAX_MAX_PAYLOAD_SIZE) {
                throw new IllegalArgumentException(
                        "The maximum payload send size must be at least 16384 bytes and at most "
                                + ChannelAccessMessageCodec.MAX_MAX_PAYLOAD_SIZE
                                + ".");
            }
            this.maxPayloadSendSize = maxPayloadSendSize;
        } else {
            initializeMaxPayloadSendSizeFromEnvironment();
        }
        alignMaxPayloadSizes();
        if (echoInterval != null) {
            if (echoInterval < 0.1 || Double.isInfinite(echoInterval)
                    || Double.isNaN(echoInterval)) {
                throw new IllegalArgumentException(
                        "The echo interval must be at least 0.1 s.");
            }
            this.echoInterval = echoInterval;
        } else {
            initializeEchoIntervalFromEnvironment();
        }
        if (defaultEventMask == null) {
            defaultEventMask = ChannelAccessEventMask.DBE_VALUE
                    .or(ChannelAccessEventMask.DBE_ALARM);
        } else {
            if (defaultEventMask.equals(ChannelAccessEventMask.DBE_NONE)) {
                throw new IllegalArgumentException(
                        "The default event mask must not be DBE_NONE.");
            }
        }
        this.defaultEventMask = defaultEventMask;
        if (cidBlockReuseTime == null) {
            cidBlockReuseTime = 900000L;
        }
        this.cidBlockReuseTime = Math.max(0, cidBlockReuseTime);
        if (errorHandler != null) {
            this.errorHandler = errorHandler;
        } else {
            this.errorHandler = new JavaUtilLoggingErrorHandler();
        }
        this.wrappedErrorHandler = new ExceptionSafeErrorHandler(
                this.errorHandler);
        if (beaconDetectorConfiguration != null) {
            this.beaconDetectorConfiguration = beaconDetectorConfiguration;
        } else {
            this.beaconDetectorConfiguration = new BeaconDetectorConfiguration(
                    this.errorHandler, null, null);
        }
        if (resolverConfiguration != null) {
            this.resolverConfiguration = resolverConfiguration;
        } else {
            this.resolverConfiguration = new ChannelNameResolverConfiguration(
                    this.charset, this.errorHandler, this.hostName,
                    this.userName, null, null, null, null, null,
                    this.echoInterval, null);
        }
    }

    private void initializeEchoIntervalFromEnvironment() {
        this.echoInterval = EnvironmentUtil
                .doubleFromEnvironment("EPICS_CA_CONN_TMO", 30.0);
        if (this.echoInterval < 0.1) {
            this.echoInterval = 0.1;
        }
    }

    private void initializeHostNameFromEnvironment() {
        this.hostName = HostNameUtil.getHostName();
        if (this.hostName == null) {
            this.hostName = "<unknown host>";
        }
    }

    private void initializeUserNameFromEnvironment() {
        this.userName = UserNameUtil.getUserName();
        if (this.userName == null) {
            this.userName = "";
        }
    }

    private int getMaxPayloadSizeFromEnvironment() {
        // If EPICS_CA_AUTO_ARRAY_BYTES is not set or true, we ignore the
        // EPICS_CA_MAX_ARRAY_BYTES environment variable and instead allow
        // payloads of arbitrary size (within the limits of this
        // implementation).
        if (EnvironmentUtil.booleanFromEnvironment("EPICS_CA_AUTO_ARRAY_BYTES",
                true)) {
            return ChannelAccessMessageCodec.MAX_MAX_PAYLOAD_SIZE;
        } else {
            return EnvironmentUtil.integerFromEnvironment(
                    "EPICS_CA_MAX_ARRAY_BYTES", 16384, 16384,
                    ChannelAccessMessageCodec.MAX_MAX_PAYLOAD_SIZE);
        }
    }

    private void initializeMaxPayloadReceiveSizeFromEnvironment() {
        this.maxPayloadReceiveSize = getMaxPayloadSizeFromEnvironment();
    }

    private void initializeMaxPayloadSendSizeFromEnvironment() {
        this.maxPayloadSendSize = getMaxPayloadSizeFromEnvironment();
    }

    private void alignMaxPayloadSizes() {
        this.maxPayloadReceiveSize = this.maxPayloadReceiveSize
                - (this.maxPayloadReceiveSize
                        % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        this.maxPayloadSendSize = this.maxPayloadSendSize
                - (this.maxPayloadSendSize
                        % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
    }

    /**
     * Returns the character set that should be used for encoding and decoding
     * strings.
     * 
     * @return character set to be used for strings.
     */
    public Charset getCharset() {
        return charset;
    }

    /**
     * Returns the host name that shall be used as the identification of the
     * local machine when connecting to a Channel Access server.
     * 
     * @return host name identifying the computer the client is running on.
     */
    public String getHostName() {
        return hostName;
    }

    /**
     * Returns the user name that shall be used as the identification of the
     * user when connecting to a Channel Access server.
     * 
     * @return user name identifying the user that is using the client.
     */
    public String getUserName() {
        return userName;
    }

    /**
     * Returns the maximum payload size for message received via TCP (in bytes).
     * The number returned is at least 16384 and always a multiple of
     * {@link ChannelAccessConstants#MESSAGE_SIZE_ALIGNMENT}.
     * 
     * @return maximum payload size for a message received via TCP.
     */
    public int getMaxPayloadReceiveSize() {
        return maxPayloadReceiveSize;
    }

    /**
     * Returns the maximum payload size for message sent via TCP (in bytes). The
     * number returned is at least 16384 and always a multiple of
     * {@link ChannelAccessConstants#MESSAGE_SIZE_ALIGNMENT}.
     * 
     * @return maximum payload size for a message sent via TCP.
     */
    public int getMaxPayloadSendSize() {
        return maxPayloadSendSize;
    }

    /**
     * Returns the default event mask. This is the event mask used when creating
     * a {@link ChannelAccessMonitor} if no event mask is specified explicitly.
     * 
     * @return default event mask to use when creating an event monitor for a
     *         channel.
     */
    public ChannelAccessEventMask getDefaultEventMask() {
        return defaultEventMask;
    }

    /**
     * Returns the time that a CID is blocked from being reused after the
     * corresponding channel has been destroyed. After destroying a channel, the
     * CID may not be reused for some time because there might still be late
     * responses to old search requests, which would be used for the wrong
     * channel if the CID was reused too early.
     * 
     * @return time a CID is blocked from reuse (in milliseconds).
     */
    public long getCIDBlockReuseTime() {
        return cidBlockReuseTime;
    }

    /**
     * <p>
     * Returns the interval between two echo requests.
     * </p>
     * 
     * <p>
     * Echo requests are sent to a Channel Access server in order to ensure that
     * the connection is still working. If a beacon packet from the same server
     * is received, this is considered proof that the network connection is okay
     * and thus no echo request is sent. For this reason, this setting should be
     * chosen to be at least two times as large as the max. interval between
     * sending two beacons.
     * </p>
     * 
     * <p>
     * When an echo request is sent and no reply is received promptly, the
     * Channel Access channels associated with the respective server are put
     * into the disconnected state. The server connection itself, however, is
     * not closed. If the connection recovers, the channels are put back into
     * the connected state. The connection is only closed by the underlying
     * operating system. However, sending an echo request can be useful in this
     * context because sending any data at all might help the operating system
     * in detecting that the TCP connection has been interrupted.
     * </p>
     * 
     * <p>
     * The minimum for this setting is 0.1 seconds.
     * </p>
     * 
     * @return echo interval (in seconds).
     */
    public double getEchoInterval() {
        return echoInterval;
    }

    /**
     * Returns the error handler that is used by all client components, unless
     * there is a component-specific configuration that defines its own error
     * handler.
     * 
     * @return error handler to be used by the client.
     */
    public ErrorHandler getErrorHandler() {
        return wrappedErrorHandler;
    }

    /**
     * Returns the configuration for the {@link BeaconDetector} used by the
     * client.
     * 
     * @return beacon detector configuration.
     */
    public BeaconDetectorConfiguration getBeaconDetectorConfiguration() {
        return beaconDetectorConfiguration;
    }

    /**
     * Returns the configuration for the {@link ChannelNameResolver} used by the
     * client.
     * 
     * @return channel-name-resolver configuration.
     */
    public ChannelNameResolverConfiguration getResolverConfiguration() {
        return resolverConfiguration;
    }

}
