/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver.internal;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.InterfaceAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.concurrent.ConcurrentLinkedQueue;

import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.beacon.BeaconDetector.BeaconAnomalyListener;
import com.aquenos.epics.jackie.client.resolver.ChannelNameResolver;
import com.aquenos.epics.jackie.client.resolver.ChannelNameResolverConfiguration;
import com.aquenos.epics.jackie.client.resolver.DefaultChannelNameResolver;
import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.CommunicationProcessor;
import com.aquenos.epics.jackie.common.io.TimerProcessor;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessSearchTCPServerMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessSearchUDPServerMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;
import com.aquenos.epics.jackie.common.util.Inet4AddressUtil;

/**
 * <p>
 * Actual implementation of the {@link DefaultChannelNameResolver}. This
 * implementation is separated so that it can be put in a private package,
 * because it has to closely interact with other classes, that should not be
 * used directly.
 * </p>
 * 
 * <p>
 * This class and the {@link RoundTripTimeEstimator}, {@link SearchTimer},
 * {@link TcpResolver}, {@link UdpResolver}, and {@link UdpSender} form one
 * component together. This component is only separated into these classes in
 * order to keep the code more compact and readable.
 * </p>
 * 
 * <p>
 * If available, this class uses a {@link BeaconDetector} to detect beacon
 * anomalies. This is mainly used to speed up the search for new channels that
 * have just become available when a new server has been started.
 * </p>
 * 
 * <p>
 * Any code that wants to instantiate a channel name resolver should instantiate
 * {@link DefaultChannelNameResolver} instead of this class.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class DefaultChannelNameResolverImpl
        implements ChannelNameResolver, CommunicationProcessor, TimerProcessor {

    /**
     * Simple queue for holding requests to add or cancel a search request.
     * 
     * @author Sebastian Marsching
     */
    private static class QueuedRequest {
        public String channelName;
        public int cid;
        public long startTime;
        public boolean cancel;
        public ChannelNameSearchResultProcessor resultProcessor;
    }

    /**
     * Simple class holding a search timer and the next time this timer should
     * be processed.
     * 
     * @author Sebastian Marsching
     */
    private static class ScheduledSearchTimer {
        long startTime;
        SearchTimer searchTimer;
    }

    private static final int MAXIMUM_NUMBER_OF_SEARCH_TIMERS = 17;

    private long autoAddressesNextUpdateTime = 0L;
    private volatile boolean beaconAnomalyDetected = false;
    private BeaconAnomalyListener beaconAnomalyListener = new BeaconAnomalyListener() {
        @Override
        public void beaconAnomalyDetected(InetSocketAddress serverAddress) {
            // This is a volatile boolean, thus we can set it without
            // synchronization. We do not have to schedule a timer because any
            // change triggered by it would not take effect before the next
            // search timer is due anyway.
            beaconAnomalyDetected = true;
        }
    };
    private int beaconAnomalyTimerIndex;
    private BeaconDetector beaconDetector;
    private HashMap<Integer, ChannelSearchRequest> cidToSearchRequest = new HashMap<Integer, ChannelSearchRequest>();
    private CommunicationController communicationController;
    private ChannelNameResolverConfiguration configuration;
    private volatile boolean destroyed = false;
    private ConcurrentLinkedQueue<QueuedRequest> newRequests = new ConcurrentLinkedQueue<QueuedRequest>();
    private RoundTripTimeEstimator rttEstimator = new RoundTripTimeEstimator();
    private HashSet<ChannelSearchRequest> scheduledSearchRequests = new HashSet<ChannelSearchRequest>();
    private PriorityQueue<ScheduledSearchTimer> searchTimerQueue = new PriorityQueue<DefaultChannelNameResolverImpl.ScheduledSearchTimer>(
            11, new Comparator<ScheduledSearchTimer>() {
                @Override
                public int compare(ScheduledSearchTimer timer1,
                        ScheduledSearchTimer timer2) {
                    if (timer1.startTime > timer2.startTime) {
                        return 1;
                    } else if (timer1.startTime < timer2.startTime) {
                        return -1;
                    } else {
                        return 0;
                    }
                }
            });
    private ArrayList<SearchTimer> searchTimers;
    private TcpResolver tcpResolver;
    private HashSet<InetSocketAddress> udpAddresses = new HashSet<InetSocketAddress>();
    private UdpResolver udpResolver;

    /**
     * <p>
     * Creates an instance of the channel name resolver's default
     * implementation. This constructor is only intended for use by
     * {@link DefaultChannelNameResolver}. Code wanting to use this resolver
     * should instantiate that class instead.
     * </p>
     * 
     * <p>
     * This constructor initializes the TCP and UDP resolvers, the search timers
     * and the round-trip-time estimator. However, some initialization is is
     * left to the {@link #setCommunicationController(CommunicationController)}
     * method, because it can only be done when the communication controller is
     * known.
     * </p>
     * 
     * <p>
     * The
     * {@link #resolveChannelName(String, int, ChannelNameSearchResultProcessor)}
     * method can safely be used immediately. The other methods only work
     * correctly once initialization has finished completely.
     * </p>
     * 
     * @param configuration
     *            configuration used by this channel name resolver.
     * @param beaconDetector
     *            beacon detector used by this channel name resolver or
     *            <code>null</code> if no beacon detector should be used. The
     *            channel name resolver uses the beacon detector for two
     *            purposes: First, when a beacon anomaly has occurred, it tries
     *            to resolve channel names that it could not resolve in the past
     *            at a higher interval. Second, when an irregular beacon is
     *            received for a server that it uses for name resolution via
     *            TCP, it checks whether the connection is still working.
     */
    public DefaultChannelNameResolverImpl(
            ChannelNameResolverConfiguration configuration,
            BeaconDetector beaconDetector) {
        assert (configuration != null);
        this.beaconDetector = beaconDetector;
        if (this.beaconDetector != null) {
            this.beaconDetector.addBeaconAnomalyListener(beaconAnomalyListener);
        }
        this.configuration = configuration;
        this.tcpResolver = new TcpResolver(this.configuration, this,
                this.beaconDetector);
        try {
            this.udpResolver = new UdpResolver(this.configuration, this);
        } catch (IOException e) {
            throw new RuntimeException(
                    "Could not initialize the sockets for the channel name resolver.",
                    e);
        }
        int intNumberOfSearchTimers;
        double doubleNumberOfSearchTimers = Math.log(
                configuration.getMaximumSearchPeriod() / 0.032) / Math.log(2.0)
                + 1.0;
        if (doubleNumberOfSearchTimers > (double) MAXIMUM_NUMBER_OF_SEARCH_TIMERS) {
            intNumberOfSearchTimers = MAXIMUM_NUMBER_OF_SEARCH_TIMERS;
        } else {
            intNumberOfSearchTimers = (int) Math
                    .floor(doubleNumberOfSearchTimers);
        }
        // The C library calculates the beacon-anomaly timer index as
        // log (beaconAnomalySearchPeriod / minRoundTripEstimate) / log(2) + 1.
        // For the (fixed) values of beaconAnomalySearchPeriod = 5 and
        // minRoundTripEstimate = 0.032, this results in the number 8.29.
        this.beaconAnomalyTimerIndex = Math.min(8, intNumberOfSearchTimers - 1);
        this.searchTimers = new ArrayList<SearchTimer>(intNumberOfSearchTimers);
        for (int timerIndex = 0; timerIndex < intNumberOfSearchTimers; timerIndex++) {
            SearchTimer searchTimer = new SearchTimer(timerIndex,
                    timerIndex > this.beaconAnomalyTimerIndex, this,
                    rttEstimator);
            searchTimers.add(searchTimer);
        }
    }

    @Override
    public void resolveChannelName(String channelName, int cid,
            ChannelNameSearchResultProcessor resultProcessor) {
        resolveChannelName(channelName, cid, resultProcessor, 0L);
    }

    @Override
    public void resolveChannelName(String channelName, int cid,
            ChannelNameSearchResultProcessor resultProcessor,
            long initialDelay) {
        if (channelName == null || resultProcessor == null) {
            throw new NullPointerException();
        }
        if (destroyed) {
            throw new IllegalStateException(
                    "The channel name resolver has been destroyed.");
        }
        // We check whether the channel name can be serialized. It can be
        // serialized if its serialized length is within the limit set by the
        // maximum size of a UDP packet and the general limit of 488 bytes.
        udpResolver.verifyChannelName(channelName);
        // This method might be called from any thread. Therefore, we cannot
        // process the request directly but add it to a queue that will be
        // processed in processTimer().
        QueuedRequest request = new QueuedRequest();
        request.channelName = channelName;
        request.cid = cid;
        if (initialDelay > 0L) {
            request.startTime = System.currentTimeMillis() + initialDelay;
        } else {
            request.startTime = 0L;
        }
        request.resultProcessor = resultProcessor;
        request.cancel = false;
        newRequests.add(request);
        // We do not have to schedule a timer, because the search timers are
        // processed periodically and there is no sense in processing the
        // request earlier if it has to wait for the search timer anyway.
    }

    @Override
    public void cancelChannelName(int cid) {
        QueuedRequest request = new QueuedRequest();
        request.cid = cid;
        request.cancel = true;
        newRequests.add(request);
        // We do not have to schedule a timer, because the search timers are
        // processed periodically and there is no sense in processing the
        // request earlier because canceling the request before processing the
        // search timers is early enough.
    }

    @Override
    public void processTimer() {
        // If the resolver has been destroyed, we do not want to process timers
        // any longer.
        if (destroyed) {
            return;
        }
        // We add new search requests that have been queued by other threads to
        // our map and add them the first search timer.
        processNewRequests();
        // If a beacon anomaly was detected, we want to push the search requests
        // from the search timers with longer intervals to a timer with a
        // shorter interval;
        processBeaconAnomaly();
        // We check whether the auto address-list needs updating.
        updateAutoAddresses();
        // If the UDP resolver has not finished sending yet, we do not want to
        // add even more packets to the queue. We can simply return because the
        // UDP resolver will call scheduleImmediateTimerProcessing() once it is
        // finished. We do not make the same check for the TCP resolver because
        // for TCP queuing works slightly differently and the throttling done
        // by the search timers should be enough.
        if (udpResolver.pendingWrite()) {
            return;
        }
        // If there are no target addresses that we could send packets too,
        // there is no sense in running the search timers. If automatic address
        // discovery is enabled, we try again when the next time for updating
        // the auto address-list has been reached. Otherwise, we return for good
        // because we will not be able to do anything in the future either. If
        // we only have TCP addresses, we wait for one of the TCP servers to
        // become available. The TcpResolver will call
        // scheduleImmediateTimerProcessing when this happens.
        if (!tcpResolver.isAnyServerAvailable() && udpAddresses.isEmpty()) {
            if (configuration.isAutoDetermineAddresses()) {
                communicationController.registerTimer(this,
                        autoAddressesNextUpdateTime);
            }
            return;
        }
        // We process all search timers that are due.
        ScheduledSearchTimer scheduledSearchTimer;
        while ((scheduledSearchTimer = searchTimerQueue.peek()) != null
                && scheduledSearchTimer.startTime <= System
                        .currentTimeMillis()) {
            SearchTimer searchTimer = searchTimerQueue.poll().searchTimer;
            scheduleSearchTimer(searchTimer, searchTimer.process());
        }
    }

    private void processNewRequests() {
        QueuedRequest request;
        final SearchTimer firstTimer = searchTimers.get(0);
        long currentTime = System.currentTimeMillis();
        while ((request = newRequests.poll()) != null) {
            if (request.cancel) {
                ChannelSearchRequest searchRequest = cidToSearchRequest
                        .remove(request.cid);
                // If there is no search request for the CID, we do not have to
                // cancel anything.
                if (searchRequest == null) {
                    continue;
                }
                // A search request for the channel may have been scheduled with
                // a delay. Therefore, we have to make sure that it will not be
                // added when the time has come.
                scheduledSearchRequests.remove(searchRequest);
                SearchTimer searchTimer = searchRequest.getSearchTimer();
                if (searchTimer != null) {
                    searchTimer.remove(searchRequest);
                }
            } else {
                if (cidToSearchRequest.containsKey(request.cid)) {
                    // This can only happen if the client did something stupid
                    // and submitted two search requests with the same ID. We
                    // simply ignore the second request.
                    continue;
                }
                final ChannelSearchRequest searchRequest = new ChannelSearchRequest(
                        request.channelName, request.cid,
                        request.resultProcessor);
                // If the search should be started immediately, we add it to the
                // search timer. Otherwise, we queue it to be added later. We
                // simply use the communication controller so that we do not
                // have to maintain our own queue.
                if (request.startTime <= currentTime) {
                    cidToSearchRequest.put(searchRequest.getCID(),
                            searchRequest);
                    searchRequest.setSearchTimer(firstTimer);
                    firstTimer.add(searchRequest);
                } else {
                    scheduledSearchRequests.add(searchRequest);
                    cidToSearchRequest.put(searchRequest.getCID(),
                            searchRequest);
                    communicationController.registerTimer(new TimerProcessor() {
                        @Override
                        public void processTimer() {
                            if (destroyed) {
                                return;
                            }
                            // If the search request has been cancelled in the
                            // meantime, we do not add it.
                            if (!scheduledSearchRequests
                                    .remove(searchRequest)) {
                                return;
                            }
                            searchRequest.setSearchTimer(firstTimer);
                            firstTimer.add(searchRequest);
                        }
                    }, request.startTime);
                }
            }
        }
    }

    private void processBeaconAnomaly() {
        // We only do something if a beacon anomaly has been detected.
        if (!beaconAnomalyDetected) {
            return;
        }
        int numberOfSearchTimers = searchTimers.size();
        for (int timerIndex = beaconAnomalyTimerIndex
                + 1; timerIndex < numberOfSearchTimers; timerIndex++) {
            // For all search timers that have a higher interval than the beacon
            // anomaly timer, we move the search requests to the beacon anomaly
            // timer.
            SearchTimer timer = searchTimers.get(timerIndex);
            // We know that there is always a search timer with the
            // beaconAnomalyTimerIndex because there is a lower bound on the
            // maximum search interval.
            moveSearchRequests(beaconAnomalyTimerIndex, timer.removeAll());
        }
        // We reset the beacon anomaly flag. Clearly, there is a race condition
        // with the beacon anomaly listener, but this is okay because we have
        // not started the actual channel search yet and the beacon anomalies
        // have a fuzzy timing anyway.
        beaconAnomalyDetected = false;
    }

    private void updateAutoAddresses() {
        // We only update the address list every 15 seconds to stay efficient.
        // This should allow for sufficiently quick recovery if the computer's
        // network configuration changes.
        if (System.currentTimeMillis() < autoAddressesNextUpdateTime) {
            return;
        }

        udpAddresses.clear();

        // First, we add all addresses that have been configured manually.
        udpAddresses.addAll(configuration.getUdpAddresses());

        // We only add addresses automatically, if this feature has been
        // enabled.
        if (configuration.isAutoDetermineAddresses()) {
            boolean foundBroadcastAddress = false;
            int serverPort = configuration.getDefaultServerPort();
            try {
                Enumeration<NetworkInterface> networkInterfaceEnumeration = NetworkInterface
                        .getNetworkInterfaces();
                while (networkInterfaceEnumeration.hasMoreElements()) {
                    NetworkInterface networkInterface = networkInterfaceEnumeration
                            .nextElement();
                    // We ignore interfaces that are not up.
                    if (!networkInterface.isUp()) {
                        continue;
                    }
                    // We do not want to add the loopback interface.
                    if (networkInterface.isLoopback()) {
                        continue;
                    }
                    // An interface may have multiple addresses. In particular,
                    // there might be IPv4 and IPv6 addresses. For Channel
                    // Access, we are only interested in IPv4. If there are
                    // multiple IPv4 addresses, they most likely have the same
                    // broadcast address, but this is okay because we add to a
                    // hash set and thus duplicate addresses are only added
                    // once.
                    for (InterfaceAddress interfaceAddress : networkInterface
                            .getInterfaceAddresses()) {
                        InetAddress broadcastAddress = interfaceAddress
                                .getBroadcast();
                        if (broadcastAddress != null
                                && broadcastAddress instanceof Inet4Address
                                && !broadcastAddress.isAnyLocalAddress()) {
                            udpAddresses.add(new InetSocketAddress(
                                    broadcastAddress, serverPort));
                            foundBroadcastAddress = true;
                        }
                    }
                    // We deliberately ignore point-to-point interfaces here.
                    // The C library adds the destination address of
                    // point-to-point interfaces, but we can only get the local
                    // address via the Java API.
                }
            } catch (SocketException e) {
                // Automatic address discovery is not critical (it can just be
                // done again) thus we simply log the exception.
                configuration.getErrorHandler().handleError(
                        DefaultChannelNameResolverImpl.class, e,
                        "Error while trying to automatically discover broadcast addresses.");
            }
            // If we could not find any broadcast address, we add the loop-back
            // address. The C library exactly does the same thing.
            if (!foundBroadcastAddress) {
                udpAddresses.add(new InetSocketAddress(
                        Inet4AddressUtil.LOOPBACK_ADDRESS, serverPort));
            }
            // We update the address list every 15 seconds, so that changes in
            // the network configuration (e.g. for mobile computers) are
            // detected.
            autoAddressesNextUpdateTime = System.currentTimeMillis() + 15000L;
        } else {
            // The address list is static, so we never have to update it.
            autoAddressesNextUpdateTime = Long.MAX_VALUE;
        }
        // The UDP resolver's list of target addresses needs to be updated
        // because it is responsible for actually sending UDP packets.
        udpResolver.setTargetAddresses(udpAddresses);
    }

    @Override
    public void setCommunicationController(
            CommunicationController communicationController) {
        this.communicationController = communicationController;
        this.udpResolver.setCommunicationController(communicationController);
        this.tcpResolver.setCommunicationController(communicationController);
        // Now we can schedule our search timers. We have to do this after we
        // passed the communication controller to the UDP resolver because
        // otherwise, the UDP resolver might be used (from a timer task) before
        // it is ready. We schedule all search timers with an initial delay that
        // is equal to the minimal delay between two runs.
        long currentTime = System.currentTimeMillis();
        long delay = 32L;
        for (int timerIndex = 0; timerIndex < searchTimers
                .size(); timerIndex++) {
            scheduleSearchTimer(searchTimers.get(timerIndex),
                    currentTime + delay);
            delay *= 2L;
        }
    }

    /**
     * Handle a search response received via TCP. This method is called by
     * {@link TcpResolver} each time it receives such a message.
     * 
     * @param message
     *            search response that has been received via TCP.
     * @param serverVersion
     *            protocol version that is supported by the server.
     */
    void processSearchResponse(ChannelAccessSearchTCPServerMessage message,
            ChannelAccessVersion serverVersion) {
        processSearchResponse(message.getChannelCID(),
                message.getServerAddress(), message.getServerPort(),
                serverVersion, false, 0);
    }

    /**
     * Handle a search response received via UDP. This method is called by
     * {@link UdpResolver} each time it receives such a message.
     * 
     * @param message
     *            search response that has been received via UDP
     * @param sequenceNumberValid
     *            flag indicating whether the <code>sequenceNumber</code> is
     *            valid and should be used. The sequence number is valid if in
     *            the same UDP packet, there has been a version message
     *            preceding the search message and this version message had a
     *            sequence number that was marked as valid.
     * @param sequenceNumber
     *            sequence number for this search response. If
     *            <code>sequenceNumberValid</code> is <code>true</code>, this is
     *            the sequence number of the search request to which this reply
     *            was sent in response. This information can be used to update
     *            the round-trip-time estimate.
     */
    void processSearchResponse(ChannelAccessSearchUDPServerMessage message,
            boolean sequenceNumberValid, int sequenceNumber) {
        processSearchResponse(message.getChannelCID(),
                message.getServerAddress(), message.getServerPort(),
                message.getVersion(), sequenceNumberValid, sequenceNumber);
    }

    private void processSearchResponse(int cid, Inet4Address serverAddress,
            int serverPort, ChannelAccessVersion serverVersion,
            boolean sequenceNumberValid, int sequenceNumber) {
        ChannelSearchRequest request = cidToSearchRequest.get(cid);
        // The message might contain an invalid server port. Therefore, we use
        // the standard port if the specified port number is not within the
        // expected range.
        if (serverPort < 1 || serverPort > 65535) {
            serverPort = configuration.getDefaultServerPort();
        }
        if (request == null) {
            // This is probably a duplicate response, we simply ignore it.
            return;
        }
        // We have found the channel, so we can remove if from our map.
        cidToSearchRequest.remove(cid);
        SearchTimer searchTimer = request.getSearchTimer();
        // The searchTimer should never be null, this is just an extra
        // safe-guard.
        if (searchTimer != null) {
            // We remove the search request from the timer because it does not
            // have to be sent again.
            searchTimer.remove(request);
            // We notify the search timer of the response so that it can update
            // the round-trip-time estimate.
            if (searchTimer.processChannelFound(sequenceNumberValid,
                    sequenceNumber)) {
                // The search timer hit a success rate of 100 percent. This
                // means that the remaining search requests should be sent as
                // soon as possible.
                scheduleSearchTimerImmediately(searchTimer);
            }
        }
        // We trigger the callback that was registered with the search request.
        request.getResultProcessor().processChannelLocation(
                request.getChannelName(), cid,
                new InetSocketAddress(serverAddress, serverPort),
                serverVersion);
    }

    /**
     * Schedule a search timer to be processed at or after a specific point in
     * time. This method is only called when the search timer was just processed
     * because its timer had expired. Therefore, we do not have to check whether
     * the search timer has already been added to the queue.
     * 
     * @param searchTimer
     *            search timer that should be processed.
     * @param startTime
     *            time at which the timer should be processed the next time.
     */
    private void scheduleSearchTimer(SearchTimer searchTimer, long startTime) {
        assert (searchTimer != null);
        ScheduledSearchTimer scheduledSearchTimer = new ScheduledSearchTimer();
        scheduledSearchTimer.searchTimer = searchTimer;
        scheduledSearchTimer.startTime = startTime;
        searchTimerQueue.add(scheduledSearchTimer);
        communicationController.registerTimer(this, startTime);
    }

    /**
     * Schedule a search timer to be processed as soon as possible. This method
     * has to check whether there is already an entry for this timer in the
     * queue because it is triggered asynchronously when a packet is received.
     * 
     * @param searchTimer
     *            search timer that shall be processed as soon as possible.
     */
    private void scheduleSearchTimerImmediately(SearchTimer searchTimer) {
        for (Iterator<ScheduledSearchTimer> timerIterator = searchTimerQueue
                .iterator(); timerIterator.hasNext();) {
            ScheduledSearchTimer scheduledSearchTimer = timerIterator.next();
            if (scheduledSearchTimer.searchTimer == searchTimer) {
                timerIterator.remove();
                scheduleSearchTimer(searchTimer, 0L);
                break;
            }
        }
    }

    /**
     * Adds a search message to the current UDP packet and to the TCP queues.
     * The UDP packet is sent when {@link #send()} is called. For TCP, the
     * message might be sent immediately. This method is only intended for use
     * by {@link SearchTimer#process()}.
     * 
     * @param cid
     *            CID of the channel (from the search request).
     * @param channelName
     *            channel name.
     * @return <code>true</code> if the message was added to the current UDP
     *         packet, <code>false</code> if it could not be added because there
     *         is not enough space left in the packet. If message is not added
     *         to the UDP packet, it is also not added to the TCP queues.
     * @throws SerializedChannelNameTooLargeException
     *             if the channel name is too long to be serialized.
     */
    boolean addSearchMessage(int cid, String channelName) {
        assert (channelName != null);
        boolean added = udpResolver.addSearchMessage(cid, channelName);
        // We only send a search request via TCP if it has been added to the UDP
        // packet. If it has not been added, it will be added to the next packet
        // and then we would sent it a second time via TCP.
        if (added) {
            tcpResolver.sendSearchMessage(cid, channelName);
        }
        return added;
    }

    /**
     * Sends the current UDP packet. This sends all search messages that have
     * been queued using {@link #addSearchMessage(int, String)} since the last
     * call of this method. If no search message have been queued, no packet is
     * sent. This method is only intended for use by
     * {@link SearchTimer#process()}.
     */
    void send() {
        udpResolver.send();
    }

    /**
     * Returns the current UDP packet sequence number. This is the sequence
     * number of the packet that will be sent the next time {@link #send()} is
     * called. This method is only intended for use by
     * {@link SearchTimer#process()}.
     * 
     * @return current UDP packet sequence number.
     */
    int currentSequenceNumber() {
        return udpResolver.currentSequenceNumber();
    }

    /**
     * Moves the specified search requests to a certain search timer. The search
     * requests are not removed from their old timer, so the calling code has to
     * ensure that they are removed. This method is only intended for use by
     * {@link SearchTimer#process()}, which uses it to move search requests it
     * currently owns to a different timer.
     * 
     * @param newTimerIndex
     *            index of the new search timer. Must not be negative. If
     *            greater than the maximum search-timer index, the maximum
     *            search-timer index is used instead.
     * @param requests
     *            requests that shall be moved.
     */
    void moveSearchRequests(int newTimerIndex,
            Collection<ChannelSearchRequest> requests) {
        assert (newTimerIndex >= 0);
        if (newTimerIndex >= searchTimers.size()) {
            newTimerIndex = searchTimers.size() - 1;
        }
        SearchTimer newSearchTimer = searchTimers.get(newTimerIndex);
        for (ChannelSearchRequest request : requests) {
            request.setSearchTimer(newSearchTimer);
            newSearchTimer.add(request);
        }
    }

    /**
     * Moves the specified search requests to the search timer that shall be
     * used if a beacon anomaly is detected. The search requests are not removed
     * from their old timer, so the calling code has to ensure that they are
     * removed. This method is intended for use by {@link SearchTimer#process()}
     * , which uses it to move search requests it currently owns to the
     * beacon-anomaly timer if it detects that it actually receives responses.
     * It is also used internally to move search requests to the beacon-anomaly
     * timer when a beacon anomaly is detected.
     * 
     * @param requests
     *            search requests that shall be moved to the beacon-anomaly
     *            timer.
     */
    void moveSearchRequestsToBeaconAnomalyPeriod(
            Collection<ChannelSearchRequest> requests) {
        moveSearchRequests(beaconAnomalyTimerIndex, requests);
    }

    /**
     * Schedules processing of due timers as soon as possible. This method is
     * called by the UDP and TCP resolvers of this channel name resolver when
     * they have finished writing all pending data. The channel name resolver
     * schedules a processing of due search timers, because this might have been
     * put on hold while it was waiting for the pending write operations to
     * finish.
     */
    void scheduleImmediateTimerProcessing() {
        communicationController.registerTimer(this, 0L);
    }

    /**
     * Processes a search request that could not be send because it caused an
     * error. This method is only intended for use by
     * {@link SearchTimer#process()}.
     * 
     * @param request
     *            search request that caused an exception when trying to send
     *            it.
     * @param exception
     *            exception that was caused.
     */
    void processFailedRequest(ChannelSearchRequest request,
            Exception exception) {
        // We verify the channel names in search requests before adding them to
        // our queue. Therefore, this should never happen. We just log the
        // exception and continue.
        configuration.getErrorHandler().handleError(
                DefaultChannelNameResolverImpl.class, exception,
                "Unexpected exception while trying to send channel search request.");
    }

    /**
     * Destroys this channel name resolver. This closes the I/O channels opened
     * by this resolver. The {@link CommunicationController} that is running the
     * communication thread should be stopped before calling this method.
     * Otherwise, unexpected exceptions might be thrown in the communication
     * thread.
     */
    public void destroy() {
        // We set the destroyed flag first. This should keep other threads from
        // trying to reopen channels that we just closed.
        destroyed = true;
        // Close TCP channels.
        tcpResolver.destroy();
        // Close the UDP channel.
        udpResolver.destroy();
        // We remove our listener from the beacon detector. The listener would
        // probably be garbage-collected eventually, but it is better to be
        // explicit.
        if (beaconDetector != null) {
            beaconDetector.removeBeaconAnomalyListener(beaconAnomalyListener);
        }
    }

}
