/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.DoubleBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.LongBuffer;
import java.nio.ShortBuffer;
import java.util.LinkedList;

/**
 * <p>
 * Byte source backed by a list of byte buffers. This byte source allows for
 * accessing a sequence of buffers like they were one contiguous buffer.
 * </p>
 * 
 * <p>
 * <strong>Attention</strong>: Special care must be taken to ensure that all
 * byte buffers added to an instance of this class have the same
 * {@link ByteOrder}. Otherwise the get methods that return multi-byte
 * data-types might have unexpected results.
 * </p>
 * 
 * <p>
 * The <code>get*Array</code> methods are optimized to only use one copy
 * operation for transferring data. An extra copy operation is necessary for
 * elements that span across two of the backing {@link ByteBuffer}s, however all
 * other elements are read with only a single copy operation. If the
 * {@link ByteBuffer}'s {@link ByteOrder} is the platforms native byte order,
 * array reads can be expected to be very efficient, because complete blocks of
 * memory are copied.
 * </p>
 * 
 * <p>
 * Sub-classes might want to override {@link #release(ByteBuffer)} in order to
 * reuse byte buffers.
 * </p>
 * 
 * <p>
 * Instances of this class are not thread-safe. If multiple threads are going to
 * access an instance of this class, access to it must be synchronized. This
 * also applies to simple read operations (without advancing the position)
 * because under certain conditions the state of the backing {@link ByteBuffer}s
 * might be changed temporarily.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ByteBufferByteSource extends AbstractByteSource {

    /**
     * Interface for a command that copies data from a byte buffer into an array
     * of a specific type. This would be a perfect example for lambda
     * expressions, but we do not want to require Java 8 yet.
     * 
     * @author Sebastian Marsching
     * 
     * @param <T>
     *            type of the array into which this reader can read data.
     */
    private static interface ValueReader<T> {

        /**
         * Reads elements from <code>buffer</code> into <code>array</code>. This
         * method may change the position of the byte buffer or other
         * properties. However, it may not change the buffer's content.
         * 
         * @param destination
         *            array into which the elements should be copied.
         * @param destinationOffset
         *            offset into <code>destination</code>. The method should
         *            start inserting elements at this position.
         * @param source
         *            buffer from which the elements should be copied. This
         *            method may change this buffer's properties (e.g. the
         *            position) but not the content of the buffer.
         * @param numberOfBytes
         *            number of bytes that should be read. The calling code
         *            ensures that there are always enough remaining bytes in
         *            the buffer in order to read the specified number of
         *            elements and that the array is big enough to hold the
         *            specified number of elements (also considering the
         *            <code>destinationOffset</code>).
         */
        void readValues(T destination, int destinationOffset,
                ByteBuffer source, int numberOfElements);

    }

    private final static ValueReader<byte[]> BYTE_READER = new ValueReader<byte[]>() {
        @Override
        public void readValues(byte[] destination, int destinationOffset,
                ByteBuffer source, int numberOfBytes) {
            // The array should be big enough, so we just use an assertion here.
            assert (destination.length >= destinationOffset + numberOfBytes);
            // The buffer should have sufficient remaining bytes, thus we just
            // use an assertion here.
            assert (source.remaining() >= numberOfBytes);

            source.get(destination, destinationOffset, numberOfBytes);
        }
    };

    private final static ValueReader<double[]> DOUBLE_READER = new ValueReader<double[]>() {
        @Override
        public void readValues(double[] destination, int destinationOffset,
                ByteBuffer source, int numberOfBytes) {
            int numberOfElements = numberOfBytes / SIZE_OF_DOUBLE;

            // The array should be big enough, so we just use an assertion here.
            assert (destination.length >= destinationOffset + numberOfElements);
            // The buffer should have sufficient remaining bytes, thus we just
            // use an assertion here.
            assert (source.remaining() >= numberOfBytes);

            DoubleBuffer typedBuffer = source.asDoubleBuffer();
            typedBuffer.get(destination, destinationOffset, numberOfElements);
        }
    };

    private final static ValueReader<float[]> FLOAT_READER = new ValueReader<float[]>() {
        @Override
        public void readValues(float[] destination, int destinationOffset,
                ByteBuffer source, int numberOfBytes) {
            int numberOfElements = numberOfBytes / SIZE_OF_FLOAT;

            // The array should be big enough, so we just use an assertion here.
            assert (destination.length >= destinationOffset + numberOfElements);
            // The buffer should have sufficient remaining bytes, thus we just
            // use an assertion here.
            assert (source.remaining() >= numberOfBytes);

            FloatBuffer typedBuffer = source.asFloatBuffer();
            typedBuffer.get(destination, destinationOffset, numberOfElements);
        }
    };

    private final static ValueReader<int[]> INT_READER = new ValueReader<int[]>() {
        @Override
        public void readValues(int[] destination, int destinationOffset,
                ByteBuffer source, int numberOfBytes) {
            int numberOfElements = numberOfBytes / SIZE_OF_INT;

            // The array should be big enough, so we just use an assertion here.
            assert (destination.length >= destinationOffset + numberOfElements);
            // The buffer should have sufficient remaining bytes, thus we just
            // use an assertion here.
            assert (source.remaining() >= numberOfBytes);

            IntBuffer typedBuffer = source.asIntBuffer();
            typedBuffer.get(destination, destinationOffset, numberOfElements);
        }
    };

    private final static ValueReader<long[]> LONG_READER = new ValueReader<long[]>() {
        @Override
        public void readValues(long[] destination, int destinationOffset,
                ByteBuffer source, int numberOfBytes) {
            int numberOfElements = numberOfBytes / SIZE_OF_LONG;

            // The array should be big enough, so we just use an assertion here.
            assert (destination.length >= destinationOffset + numberOfElements);
            // The buffer should have sufficient remaining bytes, thus we just
            // use an assertion here.
            assert (source.remaining() >= numberOfBytes);

            LongBuffer typedBuffer = source.asLongBuffer();
            typedBuffer.get(destination, destinationOffset, numberOfElements);
        }
    };

    private final static ValueReader<short[]> SHORT_READER = new ValueReader<short[]>() {
        @Override
        public void readValues(short[] destination, int destinationOffset,
                ByteBuffer source, int numberOfBytes) {
            int numberOfElements = numberOfBytes / SIZE_OF_SHORT;

            // The array should be big enough, so we just use an assertion here.
            assert (destination.length >= destinationOffset + numberOfElements);
            // The buffer should have sufficient remaining bytes, thus we just
            // use an assertion here.
            assert (source.remaining() >= numberOfBytes);

            ShortBuffer typedBuffer = source.asShortBuffer();
            typedBuffer.get(destination, destinationOffset, numberOfElements);
        }
    };

    private LinkedList<ByteBuffer> nextBuffers = new LinkedList<ByteBuffer>();
    private long remainingBytes = 0;
    private boolean atomicReadInProgress = false;
    private long atomicReadOffset = 0;
    private LinkedList<Long> atomicReadOffsetStack = new LinkedList<Long>();

    /**
     * Creates a new byte source that initially does not contain any buffers.
     * This means that calling any of the read methods throw a
     * {@link BufferUnderflowException} until at least one usable buffer is
     * appended to this source.
     */
    public ByteBufferByteSource() {
    }

    /**
     * Called every time a byte buffer is released (removed from the list of
     * remaining buffers). This method can be overridden by sub-classes in order
     * to reuse the buffer (e.g. return it to a pool of buffers). This might be
     * useful when using direct buffers, which have relatively high allocation
     * costs. The default implementation does nothing.
     * 
     * @param buffer
     *            buffer that has been removed and can be released.
     */
    protected void release(ByteBuffer buffer) {
        // The default implementation is empty (the buffer will be garbage
        // collected).
    }

    /**
     * Appends a byte buffer to the list of available buffers. This increases
     * the number of bytes remaining in this byte source by the number of bytes
     * remaining in <code>buffer</code>.
     * 
     * @param buffer
     *            buffer to be appended to the internal list of buffers.
     * @return a reference to <code>this</code> object.
     * @throws IllegalStateException
     *             if adding the buffer would push the number of remaining bytes
     *             in this byte source above {@link Long#MAX_VALUE}.
     */
    public ByteBufferByteSource appendBuffer(ByteBuffer buffer) {
        nextBuffers.addLast(buffer);
        int remainingInBuffer = buffer.remaining();
        if (Long.MAX_VALUE - remainingInBuffer < remainingBytes) {
            throw new IllegalStateException(
                    "This byte source already contains " + remainingBytes
                            + " bytes. Adding a buffer with "
                            + remainingInBuffer
                            + " would push it over the limit of "
                            + Long.MAX_VALUE + " bytes.");
        }
        remainingBytes += remainingInBuffer;
        return this;
    }

    /**
     * Appends an array of byte buffers to the list of available buffers. This
     * increases the number of bytes remaining in this byte source by the sum of
     * the number of bytes remaining in each buffer in <code>buffers</code>.
     * 
     * @param buffers
     *            buffers to be appended to the internal list of buffers.
     * @return a reference to <code>this</code> object.
     * @throws IllegalStateException
     *             if adding the buffers would push the number of remaining
     *             bytes in this byte source above {@link Integer#MAX_VALUE}.
     *             None of the buffers are added in this case.
     */
    public ByteBufferByteSource appendBuffers(ByteBuffer[] buffers) {
        boolean successful = false;
        int count = 0;
        try {
            for (ByteBuffer buffer : buffers) {
                appendBuffer(buffer);
                count++;
            }
            successful = true;
        } finally {
            if (!successful) {
                // Remove the buffers that have already been added.
                while (count > 0) {
                    nextBuffers.removeLast();
                    count--;
                }
            }
        }
        return this;
    }

    /**
     * Appends an array of byte buffers to the list of available buffers. This
     * increases the number of bytes remaining in this byte source by the sum of
     * the number of bytes remaining in each buffer in <code>buffers</code>.
     * 
     * @param buffers
     *            buffers to be appended to the internal list of buffers.
     * @return a reference to <code>this</code> object.
     * @throws IllegalStateException
     *             if adding the buffer would push the number of remaining bytes
     *             in this byte source above {@link Integer#MAX_VALUE}. None of
     *             the buffers are added in this case.
     */
    public ByteBufferByteSource appendBuffers(
            Iterable<? extends ByteBuffer> buffers) {
        boolean successful = false;
        int count = 0;
        try {
            for (ByteBuffer buffer : buffers) {
                appendBuffer(buffer);
                count++;
            }
            successful = true;
        } finally {
            if (!successful) {
                // Remove the buffers that have already been added.
                while (count > 0) {
                    nextBuffers.removeLast();
                    count--;
                }
            }
        }
        return this;
    }

    /**
     * <p>
     * Removes and returns the last buffer in the queue of byte buffers backing
     * this byte source. If this is the only remaining buffer in the queue and
     * thus data has already been read from it, the buffer's position has
     * advanced accordingly and thus the number of bytes remaining has
     * decreased. If all data in the buffer has already been read, the buffer
     * has been released and thus <code>null</code> is returned.
     * </p>
     * 
     * <p>
     * The main use for this method is to remove a partially filled buffer so
     * that more data can be added to it when the data already in the buffer is
     * not sufficient to satisfy a function reading from the byte source. This
     * way, it can be avoided to accumulate partially filled buffers in the byte
     * source which would have a negative impact on memory usage and might also
     * hurt the read performance.
     * </p>
     * 
     * <p>
     * Care must be taken when adding data to the buffer returned by this
     * method. As the buffer is prepared for reading from it, the new data must
     * be added <em>behind</em> the buffer's limit and the limit must be
     * advanced accordingly.
     * </p>
     * 
     * @return the last buffer in the queue backing this byte source or
     *         <code>null</code> if the queue is empty ({@link #remaining()}
     *         returns zero).
     * @throws IllegalStateException
     *             if an atomic read operation is in progress.
     */
    public ByteBuffer removeLastBuffer() {
        if (atomicReadInProgress) {
            throw new IllegalStateException(
                    "Removing a buffer is not allowed while an atomic read operation is in progress.");
        }
        if (nextBuffers.isEmpty()) {
            return null;
        }
        ByteBuffer buffer = nextBuffers.removeLast();
        remainingBytes -= buffer.remaining();
        return buffer;
    }

    /**
     * Reads the specified number of bytes into the supplied array, converting
     * them to the correct data-type using the supplied value reader.
     * 
     * @param numberOfValues
     *            number of values to read.
     * @param elementSize
     *            size of each value (in bytes).
     * @param destination
     *            array into which to copy the read values.
     * @param destinationOffset
     *            offset into the destination array. The values are put into the
     *            array starting at this position.
     * @param valueReader
     *            value reader for the specified array type. The value reader
     *            converts the values from bytes into the target type and puts
     *            them into the array.
     */
    private <T> void readValues(int numberOfValues, int elementSize,
            T destination, int destinationOffset, ValueReader<T> valueReader) {
        // A negative length is already detected when checking the array size,
        // therefore we only use an assertion for this test here.
        assert (numberOfValues >= 0);
        // If the number of values is zero, we can simply return.
        if (numberOfValues == 0) {
            return;
        }
        // We do the check of the read size with longs because by adding and
        // multiplying numbers, we might exceed the range of ints.
        long lengthInBytes = ((long) numberOfValues) * ((long) elementSize);
        long effectiveRemainingBytes = remainingBytes - atomicReadOffset;
        if (lengthInBytes > effectiveRemainingBytes) {
            throw new BufferUnderflowException();
        }
        // The element size should never come from an external parameter,
        // therefore we only use an assertion.
        assert (elementSize >= 1);

        readValuesInternal(atomicReadOffset, numberOfValues, elementSize,
                destination, destinationOffset, valueReader);

        if (atomicReadInProgress) {
            atomicReadOffset += numberOfValues * elementSize;
        } else {
            skipInternal(numberOfValues * elementSize);
        }
    }

    private <T> void readValuesInternal(long offset, int numberOfValues,
            int elementSize, T destination, int destinationOffset,
            ValueReader<T> valueReader) {
        // We only make assertions on the input parameters. They should already
        // have been checked before.
        assert (offset >= 0);
        assert (numberOfValues >= 0);
        assert (elementSize >= 1);
        assert (destinationOffset >= 0);

        long remainingOffset = offset;
        long remainingLength = ((long) numberOfValues) * ((long) elementSize);

        ByteBuffer temporaryBuffer = null;

        for (ByteBuffer buffer : nextBuffers) {
            int remainingInBuffer = buffer.remaining();
            // We skip empty buffers.
            if (remainingInBuffer == 0) {
                continue;
            }
            // If we have a buffer that is smaller than the current offset, we
            // can skip the complete buffer and just decrease the offset.
            if (remainingInBuffer <= remainingOffset) {
                remainingOffset -= remainingInBuffer;
                continue;
            }
            // If the temporary buffer has not been initialized yet, we treat
            // this like an empty buffer (the buffer has no data).
            int remainingSpaceInTemporaryBuffer = elementSize;
            if (temporaryBuffer != null) {
                remainingSpaceInTemporaryBuffer = temporaryBuffer.remaining();
            }
            // Do we still need to complete the data in the temporary
            // buffer in order to form a complete element?
            // We can safely cast to int here because we know that
            // remainingOffset must be less than or equal to remainingInBuffer.
            int remainingUsableBytesInBuffer = remainingInBuffer
                    - (int) remainingOffset;
            if (remainingSpaceInTemporaryBuffer != elementSize) {
                int bytesToCopy = Math.min(remainingSpaceInTemporaryBuffer,
                        remainingUsableBytesInBuffer);
                // We know that buffer.position() + remainingOffset must be in
                // the integer range, because the size of a buffer is in the
                // integer range and we already checked that remainingOffset
                // <= buffer.remaining().
                copyBytes(buffer, buffer.position() + (int) remainingOffset,
                        temporaryBuffer, bytesToCopy);
                remainingOffset += bytesToCopy;
                remainingLength -= bytesToCopy;
                remainingUsableBytesInBuffer -= bytesToCopy;
                remainingSpaceInTemporaryBuffer -= bytesToCopy;
                if (remainingSpaceInTemporaryBuffer == 0) {
                    temporaryBuffer.flip();
                    valueReader.readValues(destination, destinationOffset,
                            temporaryBuffer, elementSize);
                    destinationOffset += 1;
                    temporaryBuffer.clear();
                    remainingSpaceInTemporaryBuffer = temporaryBuffer
                            .remaining();
                }
            }
            // If the remaining length is zero, we are done.
            if (remainingLength == 0) {
                return;
            }
            // If the current buffer has more remaining bytes than the element
            // size, we can read at least one element (possibly more).
            if (remainingUsableBytesInBuffer >= elementSize) {
                // We can read as many elements as still needed or as many
                // elements as are available in the current buffer - whichever
                // is less.
                int numberOfElementsToRead = (int) Math.min(remainingLength,
                        remainingUsableBytesInBuffer) / elementSize;
                int numberOfBytesToRead = numberOfElementsToRead * elementSize;
                ByteBuffer positionedBuffer = buffer.duplicate();
                // We know that buffer.position() + remainingOffset must be in
                // the integer range, because the size of a buffer is in the
                // integer range and we already checked that remainingOffset
                // <= buffer.remaining().
                positionedBuffer.position(positionedBuffer.position()
                        + (int) remainingOffset);
                valueReader.readValues(destination, destinationOffset,
                        positionedBuffer, numberOfBytesToRead);
                destinationOffset += numberOfElementsToRead;
                remainingOffset += numberOfBytesToRead;
                remainingLength -= numberOfBytesToRead;
                remainingUsableBytesInBuffer -= numberOfBytesToRead;
            }
            // If the remaining length is zero, we are done.
            if (remainingLength == 0) {
                return;
            }
            // If there is some data remaining in the buffer (not enough for a
            // complete element - otherwise we would have read it in the last
            // block - we copy it to the temporary buffer.
            if (remainingUsableBytesInBuffer != 0) {
                // We know that the temporary buffer must be empty, because
                // otherwise the available bytes would have been used to
                // complete it. Therefore, we just use an assertion here.
                assert (remainingSpaceInTemporaryBuffer == elementSize);
                // The temporary buffer might not be initialized yet, so we have
                // to check.
                if (temporaryBuffer == null) {
                    // We use the byte order of the source buffers. This is
                    // important because otherwise we might see unexpected
                    // results when reading multi-byte data-types.
                    temporaryBuffer = ByteBuffer.allocate(elementSize).order(
                            buffer.order());
                }
                // We know that buffer.position() + remainingOffset must be in
                // the integer range, because the size of a buffer is in the
                // integer range and we already checked that remainingOffset
                // <= buffer.remaining().
                copyBytes(buffer, buffer.position() + (int) remainingOffset,
                        temporaryBuffer, remainingUsableBytesInBuffer);
                // We do not have to set these variables here, because they are
                // either overwritten or discarded at the end of the iteration.
                // However, we still set the correct values here so that we have
                // a consistent state after this block, in case some additional
                // code is added at a later point in time.
                remainingOffset += remainingUsableBytesInBuffer;
                remainingLength -= remainingUsableBytesInBuffer;
                remainingSpaceInTemporaryBuffer -= remainingUsableBytesInBuffer;
                remainingUsableBytesInBuffer = 0;
            }
            // If the remaining length is zero, we are done.
            if (remainingLength == 0) {
                return;
            }
            // We have consumed all data from the current buffer but need all
            // available data from the next buffer, thus the offset has to be
            // set to zero.
            remainingOffset = 0;
        }

        // We reached the end of the loop. This means that there is not enough
        // remaining data.
        throw new BufferUnderflowException();
    }

    private void copyBytes(ByteBuffer sourceBuffer, int positionInSourceBuffer,
            ByteBuffer destinationBuffer, int numberOfBytes) {
        ByteBuffer duplicatedSourceBuffer = sourceBuffer.duplicate();
        duplicatedSourceBuffer.position(positionInSourceBuffer + numberOfBytes);
        duplicatedSourceBuffer.flip();
        duplicatedSourceBuffer.position(positionInSourceBuffer);
        destinationBuffer.put(duplicatedSourceBuffer);
    }

    @Override
    public ByteBufferByteSource skip(int length) {
        if (length < 0) {
            throw new IllegalArgumentException("Length must not be negative.");
        }
        long effectiveRemainingBytes = remainingBytes - atomicReadOffset;
        if (length > effectiveRemainingBytes) {
            throw new BufferUnderflowException();
        }
        if (atomicReadInProgress) {
            atomicReadOffset += length;
        } else {
            skipInternal(length);
        }
        return this;
    }

    private void skipInternal(long length) {
        // We only use an assertion for this check. This method is only used
        // internally and every method calling it should have checked this
        // before.
        assert (remainingBytes >= length);
        remainingBytes -= length;
        long remainingLength = length;
        while (remainingLength > 0) {
            // It is safe to access the buffer-list without checking it, because
            // we checked there are enough remaining buffers before.
            ByteBuffer nextBuffer = nextBuffers.peek();
            int remainingInNextBuffer = nextBuffer.remaining();
            if (remainingInNextBuffer > remainingLength) {
                nextBuffer.position(nextBuffer.position()
                        + (int) remainingLength);
                return;
            }
            nextBuffer.position(nextBuffer.position() + remainingInNextBuffer);
            release(nextBuffers.pop());
            remainingLength -= remainingInNextBuffer;
        }
        // We also want to remove all empty buffers at the beginning of the
        // list. Otherwise it could happen that we still have empty buffers
        // remaining in the list after we have read all data.
        while (!nextBuffers.isEmpty() && !nextBuffers.peek().hasRemaining()) {
            release(nextBuffers.pop());
        }
    }

    @Override
    public <T> T atomicGet(AtomicGetOperation<T> atomicGetOperation) {
        // If there is already an atomic get operation in progress, we push the
        // current read offset to the stack. We will need this offset if we
        // want to roll back the get operation we are starting. If there is no
        // atomic get operation in progress yet, we start one.
        if (atomicReadInProgress) {
            atomicReadOffsetStack.push(atomicReadOffset);
        } else {
            // Setting the atomicReadInProgress flag has two consequences: Gets
            // do not cause the read data to be discarded and we will start a
            // nested atomic get operation if this method is called again.
            atomicReadInProgress = true;
        }
        // Depending on whether the operation was successful, we either discard
        // the data read or we just reset the offset.
        boolean successful = false;
        try {
            T result = atomicGetOperation.get();
            successful = true;
            return result;
        } finally {
            if (successful) {
                // If the get operations were successful and this is not a
                // nested atomic get operation, we discard the data that has
                // been read. If this is a nested operation, we do not discard
                // the data and just continue with the new offset. This has the
                // same effect as if all operations in the nested operation
                // would have been performed by the parent operation.
                if (atomicReadOffsetStack.isEmpty()) {
                    skipInternal(atomicReadOffset);
                    atomicReadOffset = 0;
                    atomicReadInProgress = false;
                } else {
                    atomicReadOffsetStack.pop();
                }
            } else {
                // If the get operations were not successful, we have to roll
                // back. For non-nested atomic get operation, we just reset the
                // offset and mark the operation as finished. For a nested
                // operation, we reset the offset to the value it had before
                // starting the nested operation. This has the same effect as if
                // the nested operation never happened.
                if (atomicReadOffsetStack.isEmpty()) {
                    atomicReadOffset = 0;
                    atomicReadInProgress = false;
                } else {
                    atomicReadOffset = atomicReadOffsetStack.pop();
                }
            }
        }
    }

    @Override
    public int remaining() {
        long effectiveRemainingBytes = remainingBytes - atomicReadOffset;
        if (effectiveRemainingBytes > Integer.MAX_VALUE) {
            return Integer.MAX_VALUE;
        } else {
            return (int) effectiveRemainingBytes;
        }
    }

    private void checkDestinationArraySize(int arrayLength, int offset,
            int length) {
        if (offset < 0) {
            throw new IllegalArgumentException("Offset must not be negative.");
        }
        if (length < 0) {
            throw new IllegalArgumentException("Length must not be negative.");
        }
        if ((long) offset + (long) length > (long) arrayLength) {
            throw new IndexOutOfBoundsException();
        }
    }

    @Override
    public ByteBufferByteSource getByteArray(byte[] destination, int offset,
            int length) {
        checkDestinationArraySize(destination.length, offset, length);
        readValues(length, SIZE_OF_BYTE, destination, offset, BYTE_READER);
        return this;
    }

    @Override
    public ByteBufferByteSource getDoubleArray(double[] destination,
            int offset, int length) {
        checkDestinationArraySize(destination.length, offset, length);
        readValues(length, SIZE_OF_DOUBLE, destination, offset, DOUBLE_READER);
        return this;
    }

    @Override
    public ByteBufferByteSource getFloatArray(float[] destination, int offset,
            int length) {
        checkDestinationArraySize(destination.length, offset, length);
        readValues(length, SIZE_OF_FLOAT, destination, offset, FLOAT_READER);
        return this;
    }

    @Override
    public ByteBufferByteSource getIntArray(int[] destination, int offset,
            int length) {
        checkDestinationArraySize(destination.length, offset, length);
        readValues(length, SIZE_OF_INT, destination, offset, INT_READER);
        return this;
    }

    @Override
    public ByteBufferByteSource getLongArray(long[] destination, int offset,
            int length) {
        checkDestinationArraySize(destination.length, offset, length);
        readValues(length, SIZE_OF_LONG, destination, offset, LONG_READER);
        return this;
    }

    @Override
    public ByteBufferByteSource getShortArray(short[] destination, int offset,
            int length) {
        checkDestinationArraySize(destination.length, offset, length);
        readValues(length, SIZE_OF_SHORT, destination, offset, SHORT_READER);
        return this;
    }

}
