/*
 * Copyright 2014 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import java.nio.BufferOverflowException;
import java.nio.ByteBuffer;

/**
 * <p>
 * Sink for writing bytes (and other data types). This is an interface that can
 * be implemented by any class that can accept data and write it to some kind of
 * channel, regardless of how the data is actually written. It allows for
 * writing a block of heterogeneous data in a single "transaction", meaning that
 * either the complete block of data is written, or none of it is written.
 * </p>
 * <p>
 * If data is written outside of an explicit transaction, each put operation is
 * considered as an individual implicit transaction. If the put operation fails,
 * no data will have been written, if it succeeds, all data will have been
 * written.
 * </p>
 * <p>
 * Please note that there is no absolute guarantee about all data being written:
 * Even if a put operation has been successful, the data might not be written
 * completely due to circumstances outside the control of the byte sink. Please
 * refer to the documentation of the individual-byte sink implementation in
 * order to find out which guarantees it makes to ensure atomic write. In
 * general, an implementation should make its best effort to ensure the
 * atomicity of write operations.
 * </p>
 * <p>
 * This interface can be used in a similar way as a {@link ByteBuffer}, however,
 * it is more flexible, because it does not have to be backed by a single
 * object.
 * </p>
 * <p>
 * In general, objects implementing this interface should be considered not
 * thread-safe and if multiple threads access them, the access must be
 * synchronized.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public interface ByteSink {

    /**
     * Interface for use with the {@link ByteSink#atomicPut(AtomicPutOperation)}
     * method. This interface should be implemented by commands that want to
     * write a contiguous block of data in a consistent way.
     * 
     * @param <T>
     *            type of the operations return value. This value is returned to
     *            the code that called
     *            {@link ByteSink#atomicPut(AtomicPutOperation)}.
     * 
     * @author Sebastian Marsching
     */
    static interface AtomicPutOperation<T> {
        /**
         * Method for writing data. If this method throws an exception, all data
         * written to the byte sink within this method is discarded. If this
         * method returns successfully, all data written by this method is
         * committed. The {@link ByteSink#atomicPut(AtomicPutOperation)} may be
         * called from this method. This will start a nested transaction: If the
         * nested transaction fails (and the exception thrown by the nested
         * transaction is caught within this method), data written within the
         * nested transaction is discarded. If the nested transaction succeeds
         * but the surrounding transaction fails, the data written by the nested
         * transaction will also be discarded. Only if both the nested
         * transaction and the surrounding transaction succeed, the data from
         * the nested transaction will be committed.
         * 
         * @return the result of the put operation. This result may be void. It
         *          is subsequently returned by
         *          {@link ByteSink#atomicPut(AtomicPutOperation)}.
         */
        T put();
    }

    /**
     * Starts an atomic put operation. Either all data written within this
     * transaction is committed, or (if the {@link AtomicPutOperation#put()}
     * method of <code>atomicPutOperation</code> throws an exception) all data
     * written within the transaction is discarded. This method will throw any
     * exception thrown by {@link AtomicPutOperation#put()}. Atomic put
     * operations may be nested (see {@link AtomicPutOperation#put()} for
     * details). Despite its name, this method should not be considered to be
     * thread-safe.
     * 
     * @param atomicPutOperation
     *            put operation that shall be performed atomically.
     * @return return value of {@link AtomicPutOperation#put()}.
     * @param <T>
     *            type of the return value of the atomic operation.
     */
    <T> T atomicPut(AtomicPutOperation<? extends T> atomicPutOperation);

    /**
     * Writes a single byte to this byte sink.
     * 
     * @param b
     *            byte to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putByte(byte b);

    /**
     * Writes a complete array of bytes to this byte sink.
     * 
     * @param source
     *            array of bytes that should be completely written to this byte
     *            sink.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putByteArray(byte[] source);

    /**
     * Writes a slice of an array of bytes to this byte sink.
     * 
     * @param source
     *            array of bytes that should be partially written to this byte
     *            sink.
     * @param offset
     *            offset into <code>source</code>. Data will be read from the
     *            array starting at this index.
     * @param length
     *            number of bytes to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     * @throws IndexOutOfBoundsException
     *             if <code>offset + length</code> is greater than
     *             <code>src.length</code>. No data is written to this byte
     *             sink.
     */
    ByteSink putByteArray(byte[] source, int offset, int length);

    /**
     * Writes a single double value to this byte sink.
     * 
     * @param value
     *            value to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putDouble(double value);

    /**
     * Writes a complete array of double values to this byte sink.
     * 
     * @param source
     *            array of double values that should be completely written to
     *            this byte sink.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putDoubleArray(double[] source);

    /**
     * Writes a slice of an array of double values to this byte sink.
     * 
     * @param source
     *            array of double values that should be partially written to
     *            this byte sink.
     * @param offset
     *            offset into <code>source</code>. Data will be read from the
     *            array starting at this index.
     * @param length
     *            number of values to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     * @throws IndexOutOfBoundsException
     *             if <code>offset + length</code> is greater than
     *             <code>src.length</code>. No data is written to this byte
     *             sink.
     */
    ByteSink putDoubleArray(double[] source, int offset, int length);

    /**
     * Writes a single float value to this byte sink.
     * 
     * @param value
     *            value to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putFloat(float value);

    /**
     * Writes a complete array of float values to this byte sink.
     * 
     * @param source
     *            array of float values that should be completely written to
     *            this byte sink.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putFloatArray(float[] source);

    /**
     * Writes a slice of an array of float values to this byte sink.
     * 
     * @param source
     *            array of float values that should be partially written to this
     *            byte sink.
     * @param offset
     *            offset into <code>source</code>. Data will be read from the
     *            array starting at this index.
     * @param length
     *            number of values to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     * @throws IndexOutOfBoundsException
     *             if <code>offset + length</code> is greater than
     *             <code>src.length</code>. No data is written to this byte
     *             sink.
     */
    ByteSink putFloatArray(float[] source, int offset, int length);

    /**
     * Writes a single int value to this byte sink.
     * 
     * @param value
     *            value to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putInt(int value);

    /**
     * Writes a complete array of int values to this byte sink.
     * 
     * @param source
     *            array of int values that should be completely written to this
     *            byte sink.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putIntArray(int[] source);

    /**
     * Writes a slice of an array of int values to this byte sink.
     * 
     * @param source
     *            array of int values that should be partially written to this
     *            byte sink.
     * @param offset
     *            offset into <code>source</code>. Data will be read from the
     *            array starting at this index.
     * @param length
     *            number of values to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     * @throws IndexOutOfBoundsException
     *             if <code>offset + length</code> is greater than
     *             <code>src.length</code>. No data is written to this byte
     *             sink.
     */
    ByteSink putIntArray(int[] source, int offset, int length);

    /**
     * Writes a single short value to this byte sink.
     * 
     * @param value
     *            value to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putShort(short value);

    /**
     * Writes a complete array of short values to this byte sink.
     * 
     * @param source
     *            array of short values that should be completely written to
     *            this byte sink.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putShortArray(short[] source);

    /**
     * Writes a slice of an array of short values to this byte sink.
     * 
     * @param source
     *            array of short values that should be partially written to this
     *            byte sink.
     * @param offset
     *            offset into <code>source</code>. Data will be read from the
     *            array starting at this index.
     * @param length
     *            number of values to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     * @throws IndexOutOfBoundsException
     *             if <code>offset + length</code> is greater than
     *             <code>src.length</code>. No data is written to this byte
     *             sink.
     */
    ByteSink putShortArray(short[] source, int offset, int length);

    /**
     * Writes a single long value to this byte sink.
     * 
     * @param value
     *            value to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putLong(long value);

    /**
     * Writes a complete array of long values to this byte sink.
     * 
     * @param source
     *            array of long values that should be completely written to this
     *            byte sink.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     */
    ByteSink putLongArray(long[] source);

    /**
     * Writes a slice of an array of long values to this byte sink.
     * 
     * @param source
     *            array of long values that should be partially written to this
     *            byte sink.
     * @param offset
     *            offset into <code>source</code>. Data will be read from the
     *            array starting at this index.
     * @param length
     *            number of values to be written.
     * @return reference to this byte sink.
     * @throws BufferOverflowException
     *             if there is not enough remaining space in the byte sink. No
     *             data is written to this byte sink.
     * @throws IllegalArgumentException
     *             if trying to write more than {@link Integer#MAX_VALUE} bytes
     *             within a single transaction. No data is written to this byte
     *             sink.
     * @throws IndexOutOfBoundsException
     *             if <code>offset + length</code> is greater than
     *             <code>src.length</code>. No data is written to this byte
     *             sink.
     */
    ByteSink putLongArray(long[] source, int offset, int length);

}
