/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.UnsupportedInThisVersionException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>CA_PROTO_ACCESS_RIGHTS</code> message. This message
 * is sent from a server to a client via TCP in order to communicate the access
 * rights for a specific channel. This information might be used by the client
 * to avoid operations that would be denied due to access control rules being in
 * effect.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_ACCESS_RIGHTS
 */
public final class ChannelAccessAccessRightsMessage extends
        ChannelAccessMessage {

    private final static int ACCESS_RIGHT_READ = 1;
    private final static int ACCESS_RIGHT_WRITE = 2;

    private int cid;
    private boolean mayRead;
    private boolean mayWrite;

    /**
     * Constructor for creating a <code>CA_PROTO_ACCESS_RIGHTS</code> message.
     * 
     * @param cid
     *            the channel CID. This is the identifier used by the client to
     *            identify the channel.
     * @param mayRead
     *            if <code>true</code> the client is allowed to read from the
     *            channel.
     * @param mayWrite
     *            if <code>true</code> the client is allowed to write to the
     *            channel.
     */
    public ChannelAccessAccessRightsMessage(int cid, boolean mayRead,
            boolean mayWrite) {
        super(ChannelAccessCommand.CA_PROTO_ACCESS_RIGHTS);

        this.cid = cid;
        this.mayRead = mayRead;
        this.mayWrite = mayWrite;
    }

    private ChannelAccessAccessRightsMessage(
            ChannelAccessMessageHeader messageHeader, int channelCID,
            boolean mayRead, boolean mayWrite) {
        super(ChannelAccessCommand.CA_PROTO_ACCESS_RIGHTS, messageHeader);

        this.cid = channelCID;
        this.mayRead = mayRead;
        this.mayWrite = mayWrite;
    }

    /**
     * Returns the channel CID used by the client. This is the identifier that
     * the client uses to identify the channel. This number might be negative or
     * zero.
     * 
     * @return channel CID assigned to the channel by the client.
     */
    public int getChannelCID() {
        return cid;
    }

    /**
     * Returns a flag indicating whether the client is allowed to read from the
     * channel.
     * 
     * @return <code>true</code> if the client may read from the channel,
     *         <code>false</code> otherwise.
     */
    public boolean isMayRead() {
        return mayRead;
    }

    /**
     * Returns a flag indicating whether the client is allowed to write to the
     * channel.
     * 
     * @return <code>true</code> if the client may write to the channel,
     *         <code>false</code> otherwise.
     */
    public boolean isMayWrite() {
        return mayWrite;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_1)) {
            throw new UnsupportedInThisVersionException(version,
                    ChannelAccessCommand.CA_PROTO_ACCESS_RIGHTS);
        }
        int accessRights = 0;
        if (mayRead) {
            accessRights |= ACCESS_RIGHT_READ;
        }
        if (mayWrite) {
            accessRights |= ACCESS_RIGHT_WRITE;
        }
        serializeHeader(byteSink, 0, (short) 0, 0, cid, accessRights, version,
                maxPayloadSize);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_1)) {
            throw new UnsupportedInThisVersionException(version,
                    ChannelAccessCommand.CA_PROTO_ACCESS_RIGHTS);
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessAccessRightsMessage other = (ChannelAccessAccessRightsMessage) obj;
        return new EqualsBuilder().append(this.cid, other.cid)
                .append(this.mayRead, other.mayRead)
                .append(this.mayWrite, other.mayWrite).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode()).append(cid)
                .append(mayRead).append(mayWrite).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("cid=");
        sb.append(cid);
        sb.append(", mayRead=");
        sb.append(mayRead);
        sb.append(", mayWrite=");
        sb.append(mayWrite);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete message (including the payload).
     */
    protected static ChannelAccessAccessRightsMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_ACCESS_RIGHTS
                .getCommandNumber());
        byteSource.skip(messageHeader.getPayloadSize());
        int accessRights = messageHeader.getContextSpecific();
        boolean mayRead = (accessRights & ACCESS_RIGHT_READ) != 0;
        boolean mayWrite = (accessRights & ACCESS_RIGHT_WRITE) != 0;
        return new ChannelAccessAccessRightsMessage(messageHeader,
                messageHeader.getCID(), mayRead, mayWrite);
    }

}
