/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.charset.Charset;

import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.exception.SerializedPayloadTooLargeException;
import com.aquenos.epics.jackie.common.exception.UnsupportedInThisVersionException;
import com.aquenos.epics.jackie.common.exception.UnsupportedProtocolVersionException;
import com.aquenos.epics.jackie.common.io.ByteSink;

/**
 * Abstract base-class for all Channel Access messages. Instances of classes
 * derived from this class are immutable and thus can be considered as
 * thread-safe.
 * 
 * @author Sebastian Marsching
 */
public abstract class ChannelAccessMessage {

    /**
     * Command (message type) for this message.
     */
    protected ChannelAccessCommand command;

    /**
     * Message header that has been used to construct this message. May be
     * <code>null</code> if this message has been constructed explicitly instead
     * of deserializing it from a byte source.
     */
    protected ChannelAccessMessageHeader originalMessageHeader;

    /**
     * Constructor for creating a Channel Access message. This constructor is
     * package-private, so that only classes in this package can extend from
     * ChannelAccessMessage.
     * 
     * @param command
     *            the command for this channel access message. This defines the
     *            type of the message. This parameter must not be
     *            <code>null</code>
     */
    ChannelAccessMessage(ChannelAccessCommand command) {
        // We only use an assertion here because we have a self-contained class
        // hierarchy and therefore this parameter should always be valid.
        assert (command != null);
        this.command = command;
    }

    /**
     * Constructor for creating a Channel Access message that has been
     * deserialized from its serialized form. This constructor is
     * package-private, so that only classes in this package can extend from
     * ChannelAccessMessage.
     * 
     * @param command
     *            the command for this channel access message. This defines the
     *            type of the message. This parameter must not be
     *            <code>null</code>
     * @param messageHeader
     *            the message header that is used to for creating this message.
     *            Later calls to {@link #getOriginalMessageHeader()} will return
     *            this object. This parameter may be <code>null</code>. This
     *            will have the same effect as using the
     *            {@link #ChannelAccessMessage(ChannelAccessCommand)}
     *            constructor.
     */
    ChannelAccessMessage(ChannelAccessCommand command,
            ChannelAccessMessageHeader messageHeader) {
        // We only use an assertion here because we have a self-contained class
        // hierarchy and therefore this parameter should always be valid.
        assert (command != null);
        this.command = command;
        this.originalMessageHeader = messageHeader;
    }

    /**
     * Serializes this message to a byte sink. This method is only intended to
     * be used by {@link ChannelAccessMessageCodec}.
     * 
     * @param byteSink
     *            byte sink that is used for writing the serialized data.
     * @param version
     *            protocol version to stay compatible with. This is the protocol
     *            version of the peer the message is sent to.
     * @param maxPayloadSize
     *            maximum size of the payload.
     * @param charset
     *            encoding to use for serializing the string data (e.g. channel
     *            names, error strings, string data-values).
     * @throws SerializedPayloadTooLargeException
     *             if the payload of the message is larger than
     *             <code>maxPayloadSize</code> or is larger than 65535 bytes
     *             when a protocol version older than Channel Access version 4.9
     *             is used.
     * @throws SerializedChannelNameTooLargeException
     *             if the channel name in a <code>CA_PROTO_SEARCH</code> message
     *             is too large.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             specified by <code>version</code>.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     * @throws UnsupportedProtocolVersionException
     *             if the specified <code>version</code> is older than the
     *             oldest supported protocol version for this message. In
     *             general, all Channel Access versions since version 4.4 should
     *             be supported.
     */
    protected abstract void serialize(ByteSink byteSink,
            ChannelAccessVersion version, int maxPayloadSize, Charset charset);

    /**
     * Returns the original header that has been used to construct this message.
     * This method may only be called on messages that have been constructed by
     * decoding a message from a byte source using the respective methods in
     * {@link ChannelAccessMessageCodec}. Calling this method on a message that
     * has been constructed directly will result in an
     * {@link UnsupportedOperationException}. You can use
     * {@link #hasOriginalMessageHeader()} to check whether an original
     * message-header is available. This method is intended to be used by
     * {@link ChannelAccessErrorMessage} in order to serialize the header of the
     * message that caused the error.
     * 
     * @return the original header that was used when this message was
     *         deserialized.
     * @throws UnsupportedOperationException
     *             if this message has been constructed from scratch instead of
     *             deserializing it from a byte source.
     */
    protected ChannelAccessMessageHeader getOriginalMessageHeader() {
        if (originalMessageHeader == null) {
            throw new UnsupportedOperationException(
                    "This message has not been constructed from a serialized form.");
        }
        return originalMessageHeader;
    }

    /**
     * Verifies that this message can be serialized. This method is only
     * intended to be used by {@link ChannelAccessMessageCodec}.
     * 
     * @param version
     *            protocol version to stay compatible with. This is the protocol
     *            version of the peer the message is sent to.
     * @param maxPayloadSize
     *            maximum size of the payload.
     * @param charset
     *            encoding to use for serializing the string data (e.g. channel
     *            names, error strings, string data-values).
     * @throws SerializedPayloadTooLargeException
     *             if the payload of the message is larger than
     *             <code>maxPayloadSize</code> or is larger than 65535 bytes
     *             when a protocol version older than Channel Access version 4.9
     *             is used.
     * @throws SerializedChannelNameTooLargeException
     *             if the channel name in a <code>CA_PROTO_SEARCH</code> message
     *             is too large.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             specified by <code>version</code>.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     * @throws UnsupportedProtocolVersionException
     *             if the specified <code>version</code> is older than the
     *             oldest supported protocol version for this message. In
     *             general, all Channel Access versions since version 4.4 should
     *             be supported.
     */
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        // Most messages can always be serialized, therefore we do not perform
        // any checks by default.
    }

    /**
     * Verifies that the message header can be serialized. This includes a check
     * of the payload size.
     * 
     * @param payloadSize
     *            size of the payload. Payload sizes greater than 0xfff8 are
     *            only supported since Channel Access version 4.9.
     * @param count
     *            "count" field. The meaning depends on the message type. Counts
     *            greater than 0xffff are only supported since ChannelAccess
     *            version 4.9 and are only valid if the value of the count field
     *            is actually an element count.
     * @param version
     *            version of the Channel Access protocol. The protocol version
     *            defines the available features. In particular, message with a
     *            payload or count greater than 0xfff are only supported since
     *            Channel Access version 4.9.
     * @param maxPayloadSize
     *            the maximum payload size to use. This method ensures that the
     *            payload size is not larger than this value.
     * @throws IllegalArgumentException
     *             if <code>payloadSize</code> or <code>count</code> are
     *             negative or if <code>payloadSize</code> is not a multiple of
     *             8.
     * @throws SerializedPayloadTooLargeException
     *             if <code>payloadSize</code> is larger than
     *             <code>maxPayloadSize</code> or if <code>payloadSize</code> or
     *             <code>count</code> are larger than 0xffff and
     *             <code>version</code> is not at least
     *             {@link ChannelAccessVersion#V4_9}.
     */
    protected void verifyHeader(int payloadSize, int count,
            ChannelAccessVersion version, int maxPayloadSize) {
        // The maximum payload size must be positive. This is already enforced
        // by the ChannelAccessMessageCodec, so we only have an assertion here.
        assert (maxPayloadSize >= 0);
        // Due to internal limitations, we cannot accept payload sizes that
        // exceed Integer.MAX_VALUE, therefore we reject negative values.
        // Basically, the same applies to the count, however we cannot make this
        // check here, because in some messages the count field is abused for
        // something other than the number of elements and thus negative values
        // might actually be valid.
        // A negative payload size actually represents a payload size that is
        // greater than the maxPayloadSize (which is always positive), so we can
        // treat this as a too large payload.
        if (payloadSize < 0) {
            throw new SerializedPayloadTooLargeException(
                    "Cannot send message with payload size "
                            + (payloadSize & 0xffffffffL)
                            + " because payload size limit is set to "
                            + maxPayloadSize + ".");
        }
        // The payload size must be aligned to 8 bytes. We do not throw a
        // MessageSerializationException but an IllegalArgumentException.
        // The calling code should have taken care of this.
        if (payloadSize % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT != 0) {
            throw new IllegalArgumentException("Payload size of " + payloadSize
                    + " bytes is not aligned to 8 bytes.");
        }
        // Channel Access versions before 4.9 did not support large headers.
        // Therefore we cannot serialize messages that have a payload size or
        // count that exceeds the range of an unsigned 16-bit integer.
        if ((payloadSize >= 0xffff || count > 0xffff)
                && !version.includes(ChannelAccessVersion.V4_9)) {
            throw new SerializedPayloadTooLargeException(
                    "Cannot send message with payload size " + payloadSize
                            + " and count " + count
                            + " over Channel Access version " + version + ".");
        }
        if (payloadSize > maxPayloadSize) {
            throw new SerializedPayloadTooLargeException(
                    "Cannot send message with payload size " + payloadSize
                            + " because payload size limit is set to "
                            + maxPayloadSize + ".");
        }
    }

    /**
     * Returns <code>true</code> if and only if this message has been
     * constructed from a serialized form and the original message header is
     * available. This means that a call to {@link #getOriginalMessageHeader()}
     * will return a valid object.
     * 
     * @return <code>true</code> if the original message header is available,
     *         <code>false</code> otherwise.
     */
    protected boolean hasOriginalMessageHeader() {
        return originalMessageHeader != null;
    }

    /**
     * Serializes the header of the message. This method will automatically use
     * the large header format if needed and supported.
     * 
     * @param byteSink
     *            byte sink that is used for serializing the message header.
     * @param payloadSize
     *            size of the payload. Payload sizes greater than 0xfff8 are
     *            only supported since Channel Access version 4.9.
     * @param dataType
     *            "data-type" field. The meaning of this field depends on the
     *            message type.
     * @param count
     *            "count" field. The meaning depends on the message type. Counts
     *            greater than 0xffff are only supported since ChannelAccess
     *            version 4.9 and are only valid if the value of the count field
     *            is actually an element count.
     * @param cid
     *            "cid" field. The meaning of this field depends on the message
     *            type.
     * @param contextSpecific
     *            "available" field. The meaning of this field depends on the
     *            message type.
     * @param version
     *            version of the Channel Access protocol. The protocol version
     *            defines the available features. In particular, message with a
     *            payload or count greater than 0xfff are only supported since
     *            Channel Access version 4.9.
     * @param maxPayloadSize
     *            the maximum payload size to use. This method ensures that the
     *            payload size is not larger than this value.
     * @throws IllegalArgumentException
     *             if <code>payloadSize</code> or <code>count</code> are
     *             negative or if <code>payloadSize</code> is not a multiple of
     *             8.
     * @throws SerializedPayloadTooLargeException
     *             if <code>payloadSize</code> is larger than
     *             <code>maxPayloadSize</code> or if <code>payloadSize</code> or
     *             <code>count</code> are larger than 0xffff and
     *             <code>version</code> is not at least
     *             {@link ChannelAccessVersion#V4_9}.
     */
    protected void serializeHeader(ByteSink byteSink, int payloadSize,
            short dataType, int count, int cid, int contextSpecific,
            ChannelAccessVersion version, int maxPayloadSize) {
        // The maximum payload size must be positive. This is already enforced
        // by the ChannelAccessMessageCodec, so we only have an assertion here.
        assert (maxPayloadSize >= 0);
        // Due to internal limitations, we cannot accept payload sizes that
        // exceed Integer.MAX_VALUE, therefore we reject negative values.
        // Basically, the same applies to the count, however we cannot make this
        // check here, because in some messages the count field is abused for
        // something other than the number of elements and thus negative values
        // might actually be valid.
        // A negative payload size actually represents a payload size that is
        // greater than the maxPayloadSize (which is always positive), so we can
        // treat this as a too large payload.
        if (payloadSize < 0) {
            throw new SerializedPayloadTooLargeException(
                    "Cannot send message with payload size "
                            + (payloadSize & 0xffffffffL)
                            + " because payload size limit is set to "
                            + maxPayloadSize + ".");
        }
        // The payload size must be aligned to 8 bytes. We do not throw a
        // MessageSerializationException but an IllegalArgumentException.
        // The calling code should have taken care of this.
        if (payloadSize % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT != 0) {
            throw new IllegalArgumentException("Payload size of " + payloadSize
                    + " bytes is not aligned to 8 bytes.");
        }
        // Channel Access versions before 4.9 did not support large headers.
        // Therefore we cannot serialize messages that have a payload size or
        // count that exceeds the range of an unsigned 16-bit integer.
        boolean needLargeHeader;
        if (payloadSize >= 0xffff || count > 0xffff) {
            needLargeHeader = true;
        } else {
            needLargeHeader = false;
        }
        if (needLargeHeader && !version.includes(ChannelAccessVersion.V4_9)) {
            throw new SerializedPayloadTooLargeException(
                    "Cannot send message with payload size " + payloadSize
                            + " and count " + count
                            + " over Channel Access version " + version + ".");
        }
        if (payloadSize > maxPayloadSize) {
            throw new SerializedPayloadTooLargeException(
                    "Cannot send message with payload size " + payloadSize
                            + " because payload size limit is set to "
                            + maxPayloadSize + ".");
        }
        byteSink.putShort(command.getCommandNumber());
        if (needLargeHeader) {
            // A large header is signaled using a payload size of 0xffff and a
            // count of 0.
            byteSink.putShort((short) 0xffff);
        } else {
            byteSink.putShort((short) payloadSize);
        }
        byteSink.putShort(dataType);
        if (needLargeHeader) {
            // A large header is signaled using a payload size of 0xffff and a
            // count of 0.
            byteSink.putShort((short) 0);
        } else {
            byteSink.putShort((short) count);
        }
        byteSink.putInt(cid);
        byteSink.putInt(contextSpecific);
        if (needLargeHeader) {
            byteSink.putInt(payloadSize);
            byteSink.putInt(count);
        }
    }

    /**
     * Calculates the number of extra padding bytes that are needed to make a
     * message with the specified payload size aligned to the alignment required
     * for all Channel Access messages.
     * 
     * @param payloadSize
     *            payload size (in bytes) that is actually needed for the data.
     * @return number of extra bytes that should be added to the end of the
     *         payload in order to make it well-aligned.
     */
    protected static int calculatePaddingSize(int payloadSize) {
        if (payloadSize > 2147483640) {
            throw new IllegalArgumentException(
                    "Aligning the payload size would exceed Integer.MAX_VALUE.");
        }
        // This construct looks a bit funny, but it always calculates the
        // correct number of padding bytes for a given payload size, including a
        // payload size of zero.
        return (7 - (payloadSize + 7)
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
    }

    /**
     * Returns the command for this Channel Access message. The command defines
     * the type of the message.
     * 
     * @return command for this message (never <code>null</code>).
     */
    public ChannelAccessCommand getCommand() {
        return command;
    }

    /**
     * <p>
     * Tells whether the specified object represents the same message.
     * </p>
     * 
     * <p>
     * Two messages are considered equal if they are of the same type and have
     * the same headers and payload. For comparing the types, the actual Java
     * type and not just the Channel Access command is compared.
     * </p>
     * 
     * @param obj
     *            reference object with which this object is compared.
     * @return <code>true</code> if <code>obj</code> is a message that is of the
     *         same type and has the same headers and payload as this message.
     *         <code>false</code> if <code>obj</code> is <code>null</code> or of
     *         a different type than this message or differs in the headers or
     *         payload.
     */
    @Override
    public boolean equals(Object obj) {
        // This base implementation cannot make a complete comparison. However,
        // it can perform some basic checks so that the implementations in the
        // child classes can directly cast to the expected type.
        if (obj == null) {
            return false;
        }
        // The classes must match exactly because all messages are represented
        // by final classes.
        if (!this.getClass().equals(obj.getClass())) {
            return false;
        }
        // Now we can cast safely.
        ChannelAccessMessage other = (ChannelAccessMessage) obj;
        // We compare the command as an extra safeguard in case one class can
        // represent two different message types. We do not compare the original
        // message headers because message equality should not depend on how the
        // message has been created (deserialized or created from constructor
        // parameters).
        return this.command == other.command;
    }

    /**
     * <p>
     * Returns a hash code value for this message. This method is supported for
     * the benefit of hash tables. The hash code is calculated as a combination
     * of the hash codes of this messages headers (including the Channel Access
     * command) and its payload (if present).
     * </p>
     * 
     * <p>
     * If two messages are equal (as indicated by the return value of
     * {@link #equals(Object)}), they also have the same hash code. However, the
     * fact that two messages have the same hash code does not imply that they
     * are equal.
     * </p>
     * 
     * @return hash code suitable for use by a hash table.
     */
    @Override
    public int hashCode() {
        // This base implementation cannot calculate a hash-code for the
        // complete message. However, it can add some information that can be
        // used by the implementations in child classes. We do not include the
        // original message headers, because they are not included in the equals
        // method either.
        return command.hashCode();
    }

    /**
     * Returns a string representation of this message. The string
     * representation contains the messages command, its header fields (if
     * relevant for the command type) and its payload if present.
     * 
     * @return string representation of this message.
     */
    @Override
    public String toString() {
        // This implementation is very simple, but the child classes will
        // override this method anyway, adding their specific headers and
        // payload if present.
        // We only override this method here so that we can add a Javadoc
        // comment and do not have to do this for each of the child classes.
        return command.toString();
    }

    /**
     * Creates a string representing a byte array and adds it to a string
     * builder. The individual bytes are printed as unsigned, two-digit hex
     * numbers (e.g. 0x0a).
     * 
     * @param sb
     *            string builder to which the string representation of the byte
     *            array should be appended.
     * @param byteArray
     *            byte array that should be printed.
     */
    protected static void addByteArrayToStringBuilder(StringBuilder sb,
            byte[] byteArray) {
        if (byteArray == null) {
            sb.append("null");
            return;
        }
        sb.append('[');
        boolean first = true;
        for (byte b : byteArray) {
            if (first) {
                first = false;
            } else {
                sb.append(", ");
            }
            sb.append("0x");
            if (b < 16 && b >= 0) {
                sb.append('0');
            }
            sb.append(Integer.toHexString(b & 0xff));
        }
        sb.append(']');
    }

}
