/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>CA_PROTO_READ_SYNC</code> message. This message is
 * sent from a client to a server via TCP in order to cancel all read requests
 * that have been sent by the client but not been fulfilled yet. The server
 * replies with the same message, having the same header values (except for the
 * payload size) but no payload. Usually, the message sent by the client also
 * does not have a payload. Typically, this message is only used by old clients
 * (EPICS 3.13 and older). Newer clients might use it with old servers, as an
 * alternative to the <code>CA_PROTO_ECHO</code> message, that is not supported
 * in old Channel Access versions.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_READ_SYNC
 */
public final class ChannelAccessReadSyncMessage extends ChannelAccessMessage {

    private short dataType = 0;
    private int count = 0;
    private int cid = 0;
    private int contextSpecific = 0;
    private byte[] payload = ArrayUtils.EMPTY_BYTE_ARRAY;

    /**
     * Constructor for creating a <code>CA_PROTO_READ_SYNC</code> message. This
     * constructor should be used for messages that are sent from a client to a
     * server.
     */
    public ChannelAccessReadSyncMessage() {
        super(ChannelAccessCommand.CA_PROTO_READ_SYNC);
    }

    /**
     * Constructor for creating a <code>CA_PROTO_READ_SYNC</code> message. This
     * constructor should be used for messages that are sent from server to a
     * client in reply to a <code>CA_PROTO_READ_SYNC</code> message that has
     * been received from the client. The header fields from the supplied
     * message are used to initialize the header fields of this message.
     * 
     * @param messageFromClient
     *            the message that has been received from the client, to which
     *            this message is a reply.
     */
    public ChannelAccessReadSyncMessage(
            ChannelAccessReadSyncMessage messageFromClient) {
        super(ChannelAccessCommand.CA_PROTO_READ_SYNC);

        this.dataType = messageFromClient.dataType;
        this.count = messageFromClient.count;
        this.cid = messageFromClient.cid;
        this.contextSpecific = messageFromClient.contextSpecific;
    }

    private ChannelAccessReadSyncMessage(
            ChannelAccessMessageHeader messageHeader, short dataType,
            int count, int cid, int contextSpecific, byte[] payload) {
        super(ChannelAccessCommand.CA_PROTO_READ_SYNC, messageHeader);

        this.dataType = dataType;
        this.count = count;
        this.cid = cid;
        this.contextSpecific = contextSpecific;
        this.payload = payload;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        if (payload == null) {
            throw new UnsupportedOperationException(
                    "Cannot serialize the message payload for a message that has been constructed without the payload.");
        }
        serializeHeader(byteSink, payload.length, dataType, count, cid,
                contextSpecific, version, maxPayloadSize);
        byteSink.putByteArray(payload);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        if (payload == null) {
            throw new UnsupportedOperationException(
                    "Cannot serialize the message payload for a message that has been constructed without the payload.");
        }
        verifyHeader(payload.length, count, version, maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessReadSyncMessage other = (ChannelAccessReadSyncMessage) obj;
        return new EqualsBuilder().append(this.dataType, other.dataType)
                .append(this.count, other.count).append(this.cid, other.cid)
                .append(this.contextSpecific, other.contextSpecific)
                .append(this.payload, other.payload).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(dataType).append(count).append(cid)
                .append(contextSpecific).append(payload).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("dataType=");
        sb.append(dataType);
        sb.append(", count=");
        sb.append(count);
        sb.append(", cid=");
        sb.append(cid);
        sb.append(", contextSpecific=");
        sb.append(contextSpecific);
        sb.append(", payload=");
        addByteArrayToStringBuilder(sb, payload);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete message (including he payload if
     *             <code>headerOnly</code> is <code>false</code>).
     */
    protected static ChannelAccessReadSyncMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_READ_SYNC
                .getCommandNumber());
        byte[] payload;
        if (headerOnly) {
            payload = null;
        } else {
            payload = byteSource.getByteArray(messageHeader.getPayloadSize());
        }
        return new ChannelAccessReadSyncMessage(messageHeader,
                messageHeader.getDataType(), messageHeader.getCount(),
                messageHeader.getCID(), messageHeader.getContextSpecific(),
                payload);
    }

}
