/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.concurrent.locks.ReentrantLock;

import com.aquenos.epics.jackie.common.exception.ConcurrentNotificationException;

/**
 * <p>
 * Abstract base implementation of a {@link ListenableFuture}.
 * </p>
 * 
 * <p>
 * This implementation stores the registered listeners in a {@link HashSet} and
 * provides a {@link #notifyCompletionListeners()} method that should be invoked
 * by derived classes.
 * </p>
 * 
 * <p>
 * Derived classes can also override the
 * {@link #notifyCompletionListener(FutureCompletionListener)} method if they
 * want to modify the behavior how listeners are called.
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <V>
 *            The result type returned by the future.
 */
public abstract class AbstractListenableFuture<V>
        implements ListenableFuture<V> {

    /**
     * Registration of a completion listener. This is a separate class because
     * we have to store some flags together with the lock object.
     * 
     * @author Sebastian Marsching
     *
     * @param <V>
     *            value type of the future holding this listener registration..
     */
    private static class ListenerRegistration<V> {

        public final FutureCompletionListener<? super V> listener;
        public boolean active = true;
        public final ReentrantLock lock = new ReentrantLock();

        public ListenerRegistration(
                FutureCompletionListener<? super V> listener) {
            this.listener = listener;
        }

    }

    /**
     * Synchronization object that this class synchronizes on when retrieving or
     * modifying the set of registered listeners.
     */
    protected final Object completionListenerLock = new Object();

    /**
     * Lock policy used for completion listeners. The lock policy defines what
     * should be done when the
     * {@link #removeCompletionListener(FutureCompletionListener)} listener
     * method is called while a notification for the listener that shall be
     * removed is in progress.
     */
    protected final ListenerLockPolicy listenerLockPolicy;

    // We store the listener registrations in a map instead of a set. This way,
    // we can efficiently remove a registration for a specific listener.
    private HashMap<FutureCompletionListener<? super V>, ListenerRegistration<V>> listeners = new HashMap<FutureCompletionListener<? super V>, ListenerRegistration<V>>();
    private boolean notificationDone;

    /**
     * Default constructor. This constructor can be used by classes that inherit
     * from this class. It initializes the future with an empty set of listeners
     * and a listener lock-policy of {@link ListenerLockPolicy#BLOCK BLOCK}.
     * Child classes that want to set the {@link ListenerLockPolicy} should use
     * the {@link #AbstractListenableFuture(ListenerLockPolicy)} constructor
     * instead.
     */
    protected AbstractListenableFuture() {
        this(ListenerLockPolicy.BLOCK);
    }

    /**
     * Initializes this class with the specified listener lock-policy and an
     * empty set of listeners.
     * 
     * @param listenerLockPolicy
     *            listener lock-policy used by this future. The lock policy
     *            defines the behavior of the
     *            {@link #removeCompletionListener(FutureCompletionListener)}
     *            method.
     * @since EPICS Jackie 2.0.0
     */
    protected AbstractListenableFuture(ListenerLockPolicy listenerLockPolicy) {
        if (listenerLockPolicy == null) {
            throw new NullPointerException(
                    "The listener lock-policy must not be null.");
        }
        this.listenerLockPolicy = listenerLockPolicy;
    }

    /**
     * Notifies all currently registered completion listeners. This method
     * should be called exactly once, after the future's result became
     * available. It delegates calling each listener to
     * {@link #notifyCompletionListener(FutureCompletionListener)}, so that
     * derived classes can take extra precautions (e.g. catch certain exceptions
     * thrown by a listener).
     * 
     * @throws IllegalStateException
     *             if this method has already been called and thus the listeners
     *             have already been notified or if this future has not
     *             completed yet ({@link #isDone()} returns <code>false</code>).
     */
    protected void notifyCompletionListeners() {
        ArrayList<ListenerRegistration<V>> localListeners;
        synchronized (completionListenerLock) {
            if (notificationDone) {
                throw new IllegalStateException(
                        "notifyCompletionListeners() has already been called.");
            }
            if (!isDone()) {
                throw new IllegalStateException(
                        "The notifyCompletionListeners() method should only be called if the future has completed (isDone() returns true).");
            }
            // We set the notification done flag before calling the listeners.
            // The flag should be set even if one of the listeners throws an
            // exception.
            notificationDone = true;
            // We make a copy of the current set of listeners. This way, we do
            // not have to hold the connectionListenerLock when calling the
            // listeners.
            localListeners = new ArrayList<ListenerRegistration<V>>(
                    listeners.values());
        }
        // Now, that we have released the completionListenerLock, we can notify
        // the listeners.
        for (ListenerRegistration<V> listenerRegistration : localListeners) {
            // We have to synchronize on the per-listener lock while checking
            // the listener's registration and also keep the lock while
            // notifying the listener. This way, the listener cannot be removed
            // while we notify it and is never notified after it has been
            // removed.
            listenerRegistration.lock.lock();
            try {
                if (listenerRegistration.active) {
                    notifyCompletionListener(listenerRegistration.listener);
                }
            } finally {
                listenerRegistration.lock.unlock();
            }
        }
    }

    /**
     * Notifies a completion listener. This method can be overridden in order to
     * modify how the listener is called. For example, some implementations
     * might want to catch certain exceptions thrown by a listener.
     * 
     * @param listener
     *            the listener that should be notified. Notification should
     *            happen by calling the listeners
     *            {@link FutureCompletionListener#completed(ListenableFuture)
     *            completed()} method, passing a reference to this future
     *            instance.
     */
    protected void notifyCompletionListener(
            FutureCompletionListener<? super V> listener) {
        listener.completed(this);
    }

    @Override
    public boolean addCompletionListener(
            FutureCompletionListener<? super V> listener) {
        // We do not allow adding null listeners, because this does not make any
        // sense and would only lead to problems when trying to notify the
        // listeners.
        if (listener == null) {
            throw new NullPointerException(
                    "The listener parameter must not be null.");
        }
        // We both check whether the future has completed and add the listener
        // while holding the lock, so that it only gets added if the future has
        // not completed yet. This ensures that the listener will be called
        // exactly once - either from this method of when the future completes.
        boolean localNotificationDone;
        ListenerRegistration<V> listenerRegistration;
        synchronized (completionListenerLock) {
            // If the listener has already been added, we do not do anything and
            // simply return.
            if (listeners.containsKey(listener)) {
                return false;
            }
            listenerRegistration = new ListenerRegistration<V>(listener);
            listeners.put(listener, listenerRegistration);
            localNotificationDone = notificationDone;
        }
        // If the notification has already been done (and we have to notify the
        // listener), we do this after releasing the lock. However, we acquire
        // the per-listener lock. This way, if a another thread removes the
        // listener while we are still in this method, the listener will not get
        // removed before the notification has finished.
        // If we cannot acquire the lock, we simply continue. This is safe
        // because there is only one reason why the lock might be held by a
        // different thread (after the notificationDone flag has been set):
        // The listener must be in the process of being removed. In this case,
        // there is no need for us to call the listener now.
        if (localNotificationDone) {
            boolean gotLock = listenerRegistration.lock.tryLock();
            if (gotLock) {
                try {
                    // The listener might have been removed before we acquired
                    // the lock, so we have to check the active flag.
                    if (listenerRegistration.active) {
                        notifyCompletionListener(listener);
                    }
                } finally {
                    listenerRegistration.lock.unlock();
                }
            }
        }
        return true;
    }

    @Override
    public ListenerLockPolicy getListenerLockPolicy() {
        return listenerLockPolicy;
    }

    @Override
    public boolean removeCompletionListener(
            FutureCompletionListener<? super V> listener) {
        ListenerRegistration<V> listenerRegistration;
        synchronized (completionListenerLock) {
            listenerRegistration = listeners.remove(listener);
        }
        // We acquire the per-listener lock after releasing the
        // completionListenerLock. This way, we can avoid holding two locks at
        // the same time.
        if (listenerRegistration != null) {
            boolean gotLock;
            if (listenerLockPolicy == ListenerLockPolicy.BLOCK) {
                listenerRegistration.lock.lock();
                gotLock = true;
            } else {
                gotLock = listenerRegistration.lock.tryLock();
            }
            if (gotLock) {
                try {
                    listenerRegistration.active = false;
                } finally {
                    listenerRegistration.lock.unlock();
                }
            } else {
                if (listenerLockPolicy == ListenerLockPolicy.REPORT) {
                    throw new ConcurrentNotificationException(
                            "The listener has been removed but a notification is already in progress, so the listener is still going to be notified.");
                }
            }
            return true;
        } else {
            return false;
        }
    }

}
