/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import java.nio.IntBuffer;

/**
 * <p>
 * Channel Access value that stores integer numbers. For historical reasons,
 * this type is called <code>long</code> in the Channel Access protocol.
 * However, it is represented by <code>int</code>s.
 * </p>
 * 
 * <p>
 * This interface should not be implemented by user-defined classes. Code that
 * wants to create an instance implementing this interface should use the
 * methods provided by {@link ChannelAccessValueFactory} instead.
 * </p>
 * 
 * @author Sebatian Marsching
 */
public interface ChannelAccessLong extends ChannelAccessGettableValue<Integer> {

    /**
     * <p>
     * Returns the buffer that stores the actual value elements.
     * </p>
     * 
     * <p>
     * The buffer returned is guaranteed to have a position of zero and a
     * capacity and remaining elements equal to the number of elements in this
     * value. In addition to that, the length of the array backing this buffer
     * is equal to the buffer's capacity.
     * </p>
     * 
     * <p>
     * If this value is not read-only, the returned buffer's content can be
     * modified and this modifications will directly affect this value and vice
     * versa. Therefore, the buffer may not be modified after this value has
     * been passed to a different component.
     * </p>
     * 
     * <p>
     * If this value is read-only, the returned buffer is also read-only.
     * </p>
     * 
     * @return buffer storing this value's elements (never <code>null</code>).
     */
    IntBuffer getValue();

    /**
     * <p>
     * Replaces the buffer that stores the value's actual elements with a buffer
     * wrapping the passed array.
     * </p>
     * 
     * <p>
     * The buffer is backed by the passed array. This means that modifications
     * of the array result in modifications of this value's elements and
     * vice-versa.
     * </p>
     * 
     * <p>
     * The array passed to this method must not be changed after this value has
     * been passed to another component, because changes to the array's content
     * will affect this value's elements.
     * </p>
     * 
     * @param value
     *            array that replaces this value's elements.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setValue(int[] value);

    /**
     * <p>
     * Returns a read-only version of this value. If the value is read-only,
     * this value is returned. Otherwise, a wrapped version of this value, that
     * does not allow modifications is returned.
     * </p>
     * 
     * <p>
     * Please note that the read-only value is not a copy of this value, but
     * just a wrapper around it. This means, that modifications of the wrapped
     * value will result in modifications of the read-only value. Therefore it
     * is bad practice to modify a value that has been used as the base for a
     * read-only value, because code that still has a reference to the read-only
     * value might not expect this change and thus fail unexpectedly.
     * </p>
     * 
     * @return read-only version of this value (never <code>null</code>).
     */
    ChannelAccessLong asReadOnlyValue();

    /**
     * <p>
     * Creates and returns a copy of this object. The object returned is
     * completely independent from this object. This means that modifications to
     * this object will not affect the returned object and vice-versa.
     * </p>
     * 
     * <p>
     * The object returned is guaranteed to be completely identical to this
     * object, including its type. This means that
     * <code>x.clone().equals(x)</code> and
     * <code>x.clone().getClass() == x.getClass()</code> are always
     * <code>true</code>.
     * </p>
     * 
     * <p>
     * The returned value is guaranteed to allow write access, even if this
     * value is read-only. This means that cloning a read-only value is a good
     * way to get a writable copy.
     * </p>
     * 
     * @return a clone of this instance.
     */
    ChannelAccessLong clone();

}
