/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

/**
 * <p>
 * Channel Access values that represent a type that stores numbers and provides
 * alarm information, engineering units, and alarm and display limits.
 * </p>
 * 
 * <p>
 * This interface should not be implemented by user-defined classes. Code that
 * wants to create an instance implementing this interface should use the
 * methods provided by {@link ChannelAccessValueFactory} instead.
 * </p>
 * 
 * @param <ElementType>
 *            type of the actual value elements. For primitives, this is the
 *            boxed variant.
 * 
 * @author Sebatian Marsching
 */
public interface ChannelAccessNumericGraphicsValue<ElementType extends Number>
        extends ChannelAccessGraphicsValue<ElementType> {

    /**
     * Returns the engineerings units for this value.
     * 
     * @return engineering units (never <code>null</code> but possibly the empty
     *         string).
     * @see #getRawUnits()
     * @see #getCharset()
     */
    String getUnits();

    /**
     * <p>
     * Returns the engineering units for this value in their raw representation.
     * The raw representation always is an array of exactly eight bytes and
     * stores a null-terminated string.
     * </p>
     * <p>
     * This method returns a copy of the internal array, so changes to the
     * returned array will not affect this value. If changes shall take effect,
     * the {@link #setRawUnits(byte[])} method has to be called.
     * </p>
     * 
     * @return engineering units in their raw representation.
     * @see #getUnits()
     * @see #getCharset()
     */
    byte[] getRawUnits();

    /**
     * Sets the engineerings units for this value. This sets both the
     * {@link String Java string} and the raw representation. If the raw
     * representation using the specified {@link #getCharset() encoding} exceeds
     * the limit of eight characters (including the terminating null-byte), it
     * is silently trimmed to fit within this limit.
     * 
     * @param units
     *            new engineering units (never <code>null</code> but possibly
     *            the empty string).
     * @see #setRawUnits(byte[])
     * @see #getCharset()
     */
    void setUnits(String units);

    /**
     * <p>
     * Sets the engineerings units for this value. This sets both the the raw
     * representation and {@link String Java string}. The Java string is
     * generated by converting the specified raw representation using the
     * specified {@link #getCharset() encoding}.
     * </p>
     * <p>
     * The supplied array is copied and only the first seven bytes are used (the
     * eighth byte is reserved for the terminating null-byte). If the supplied
     * array has less than seven bytes, the remaining bytes will be filled with
     * zero.
     * </p>
     * 
     * @param rawUnits
     *            byte array containing the new raw representation (never null).
     *            Only the first seven bytes of the supplied array are used.
     * @see #setUnits(String)
     * @see #getCharset()
     */
    void setRawUnits(byte[] rawUnits);

    /**
     * <p>
     * Returns the upper display limit. The upper display limit is the limit
     * that the value does typically not exceed and thus is suitable for
     * choosing the layout of graphic displays (e.g. gauges). A value of zero
     * typically indicates that the limit has not been configured and thus
     * should not be used.
     * </p>
     * 
     * <p>
     * Code that wants to get this limit should prefer the type-specific methods
     * that return primitives for better performance. This method is only
     * provided for code that has to deal with values in a generic way (without
     * knowing the exact type of the actual value).
     * </p>
     * 
     * @return upper display limit (never <code>null</code>).
     */
    ElementType getGenericUpperDisplayLimit();

    /**
     * <p>
     * Returns the lower display limit. The lower display limit is the limit
     * that the value does typically not undershoot and thus is suitable for
     * choosing the layout of graphic displays (e.g. gauges). A value of zero
     * typically indicates that the limit has not been configured and thus
     * should not be used.
     * </p>
     * 
     * <p>
     * Code that wants to get this limit should prefer the type-specific methods
     * that return primitives for better performance. This method is only
     * provided for code that has to deal with values in a generic way (without
     * knowing the exact type of the actual value).
     * </p>
     * 
     * @return lower display limit (never <code>null</code>).
     */
    ElementType getGenericLowerDisplayLimit();

    /**
     * <p>
     * Returns the upper alarm limit. The upper alarm limit is the limit that,
     * when exceeded, typically raises a major alarm. A value of zero typically
     * indicates that the limit has not been configured and thus should not be
     * used.
     * </p>
     * 
     * <p>
     * Code that wants to get this limit should prefer the type-specific methods
     * that return primitives for better performance. This method is only
     * provided for code that has to deal with values in a generic way (without
     * knowing the exact type of the actual value).
     * </p>
     * 
     * @return upper alarm limit (never <code>null</code>).
     */
    ElementType getGenericUpperAlarmLimit();

    /**
     * <p>
     * Returns the upper warning limit. The upper warning limit is the limit
     * that, when exceeded, typically raises a minor alarm. A value of zero
     * typically indicates that the limit has not been configured and thus
     * should not be used.
     * </p>
     * 
     * <p>
     * Code that wants to get this limit should prefer the type-specific methods
     * that return primitives for better performance. This method is only
     * provided for code that has to deal with values in a generic way (without
     * knowing the exact type of the actual value).
     * </p>
     * 
     * @return upper warning limit (never <code>null</code>).
     */
    ElementType getGenericUpperWarningLimit();

    /**
     * <p>
     * Returns the lower warning limit. The lower warning limit is the limit
     * that, when undershot, typically raises a minor alarm. A value of zero
     * typically indicates that the limit has not been configured and thus
     * should not be used.
     * </p>
     * 
     * <p>
     * Code that wants to get this limit should prefer the type-specific methods
     * that return primitives for better performance. This method is only
     * provided for code that has to deal with values in a generic way (without
     * knowing the exact type of the actual value).
     * </p>
     * 
     * @return lower warning limit (never <code>null</code>).
     */
    ElementType getGenericLowerWarningLimit();

    /**
     * <p>
     * Returns the lower alarm limit. The lower alarm limit is the limit that,
     * when undershot, typically raises a major alarm. A value of zero typically
     * indicates that the limit has not been configured and thus should not be
     * used.
     * </p>
     * 
     * <p>
     * Code that wants to get this limit should prefer the type-specific methods
     * that return primitives for better performance. This method is only
     * provided for code that has to deal with values in a generic way (without
     * knowing the exact type of the actual value).
     * </p>
     * 
     * @return lower alarm limit (never <code>null</code>).
     */
    ElementType getGenericLowerAlarmLimit();

    /**
     * <p>
     * Returns a read-only version of this value. If the value is read-only,
     * this value is returned. Otherwise, a wrapped version of this value, that
     * does not allow modifications is returned.
     * </p>
     * 
     * <p>
     * Please note that the read-only value is not a copy of this value, but
     * just a wrapper around it. This means, that modifications of the wrapped
     * value will result in modifications of the read-only value. Therefore it
     * is bad practice to modify a value that has been used as the base for a
     * read-only value, because code that still has a reference to the read-only
     * value might not expect this change and thus fail unexpectedly.
     * </p>
     * 
     * @return read-only version of this value (never <code>null</code>).
     */
    ChannelAccessNumericGraphicsValue<ElementType> asReadOnlyValue();

    /**
     * <p>
     * Creates and returns a copy of this object. The object returned is
     * completely independent from this object. This means that modifications to
     * this object will not affect the returned object and vice-versa.
     * </p>
     * 
     * <p>
     * The object returned is guaranteed to be completely identical to this
     * object, including its type. This means that
     * <code>x.clone().equals(x)</code> and
     * <code>x.clone().getClass() == x.getClass()</code> are always
     * <code>true</code>.
     * </p>
     * 
     * <p>
     * The returned value is guaranteed to allow write access, even if this
     * value is read-only. This means that cloning a read-only value is a good
     * way to get a writable copy.
     * </p>
     * 
     * @return a clone of this instance.
     */
    ChannelAccessNumericGraphicsValue<ElementType> clone();

}
