/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import java.nio.charset.Charset;

/**
 * Identifier for Channel Access data-types. Each of these identifiers uniquely
 * identifies a specific data type used to transfer data via Channel Access. An
 * instance of a specific value type can be created using the
 * {@link ChannelAccessValueFactory#createValue(ChannelAccessValueType, Charset)}
 * method.
 *
 * @see ChannelAccessValue
 * 
 * @author Sebastian Marsching
 */
public enum ChannelAccessValueType {

    /**
     * Type representing a simple string value.
     * 
     * @see ChannelAccessSimpleOnlyString
     */
    DBR_STRING(0),

    /**
     * Type representing a simple short integer value.
     * 
     * @see ChannelAccessSimpleOnlyShort
     */
    DBR_SHORT(1),

    /**
     * Type representing a simple single-precision floating-point value.
     * 
     * @see ChannelAccessSimpleOnlyFloat
     */
    DBR_FLOAT(2),

    /**
     * Type representing a simple enum value. Internally, an enum is simply a
     * short, but it is interpreted as a set of states to choose from.
     * 
     * @see ChannelAccessSimpleOnlyEnum
     */
    DBR_ENUM(3),

    /**
     * Type representing a simple byte value. For historical reasons, this type
     * is called <code>DBR_CHAR</code>, however it represents an 8-bit integer (
     * <code>byte</code> in Java). Channel Access specifies that this type is
     * unsigned, but due to limitations of the Java platform, we us the signed
     * version.
     * 
     * @see ChannelAccessSimpleOnlyChar
     */
    DBR_CHAR(4),

    /**
     * Type representing a simple integer value. For historical reasons, this
     * type is called <code>DBR_LONG</code>, however it represents a 32-bit
     * integer (<code>int</code> in Java).
     * 
     * @see ChannelAccessSimpleOnlyLong
     */
    DBR_LONG(5),

    /**
     * Type representing a simple double-precision floating-point value.
     * 
     * @see ChannelAccessSimpleOnlyDouble
     */
    DBR_DOUBLE(6),

    /**
     * Type representing a string value with alarm information.
     * 
     * @see ChannelAccessAlarmOnlyString
     */
    DBR_STS_STRING(7),

    /**
     * Type representing a short integer value with alarm information.
     * 
     * @see ChannelAccessAlarmOnlyShort
     */
    DBR_STS_SHORT(8),

    /**
     * Type representing a single-precision floating-point value with alarm
     * information.
     * 
     * @see ChannelAccessAlarmOnlyFloat
     */
    DBR_STS_FLOAT(9),

    /**
     * Type representing an enum value with alarm information. Internally, an
     * enum is simply a short, but it is interpreted as a set of states to
     * choose from.
     * 
     * @see ChannelAccessAlarmOnlyEnum
     */
    DBR_STS_ENUM(10),

    /**
     * Type representing a simple byte value with alarm information. For
     * historical reasons, this type is called <code>DBR_STS_CHAR</code>,
     * however it represents an 8-bit integer (<code>byte</code> in Java).
     * Channel Access specifies that this type is unsigned, but due to
     * limitations of the Java platform, we us the signed version.
     * 
     * @see ChannelAccessAlarmOnlyChar
     */
    DBR_STS_CHAR(11),

    /**
     * Type representing an integer value with alarm information. For historical
     * reasons, this type is called <code>DBR_STS_LONG</code>, however it
     * represents a 32-bit integer (<code>int</code> in Java).
     * 
     * @see ChannelAccessAlarmOnlyLong
     */
    DBR_STS_LONG(12),

    /**
     * Type representing a double-precision floating-point value with alarm
     * information.
     * 
     * @see ChannelAccessAlarmOnlyDouble
     */
    DBR_STS_DOUBLE(13),

    /**
     * Type representing a string value with alarm information and a time-stamp.
     * 
     * @see ChannelAccessTimeString
     */
    DBR_TIME_STRING(14),

    /**
     * Type representing a short integer value with alarm information and a
     * time-stamp.
     * 
     * @see ChannelAccessTimeShort
     */
    DBR_TIME_SHORT(15),

    /**
     * Type representing a single-precision floating-point value with alarm
     * information and a time-stamp.
     * 
     * @see ChannelAccessTimeFloat
     */
    DBR_TIME_FLOAT(16),

    /**
     * Type representing an enum value with alarm information and a time-stamp.
     * Internally, an enum is simply a short, but it is interpreted as a set of
     * states to choose from.
     * 
     * @see ChannelAccessTimeEnum
     */
    DBR_TIME_ENUM(17),

    /**
     * Type representing a simple byte value with alarm information and a
     * time-stamp. For historical reasons, this type is called
     * <code>DBR_TIME_CHAR</code>, however it represents an 8-bit integer (
     * <code>byte</code> in Java). Channel Access specifies that this type is
     * unsigned, but due to limitations of the Java platform, we us the signed
     * version.
     * 
     * @see ChannelAccessTimeChar
     */
    DBR_TIME_CHAR(18),

    /**
     * Type representing an integer value with alarm information and a
     * time-stamp. For historical reasons, this type is called
     * <code>DBR_TIME_LONG</code>, however it represents a 32-bit integer (
     * <code>int</code> in Java).
     * 
     * @see ChannelAccessTimeLong
     */
    DBR_TIME_LONG(19),

    /**
     * Type representing a double-precision floating-point value with alarm
     * information and a time-stamp.
     * 
     * @see ChannelAccessTimeDouble
     */
    DBR_TIME_DOUBLE(20),

    /**
     * Type representing a string value with alarm information. Unlike other
     * data-types, strings do not have any additional graphics information, so
     * the representation of this type is identical to the one used by
     * {@link #DBR_STS_STRING}.
     * 
     * @see ChannelAccessAlarmOnlyString
     */
    DBR_GR_STRING(21),

    /**
     * Type representing a short integer value with alarm information,
     * engineering units, and alarm and display limits.
     * 
     * @see ChannelAccessGraphicsOnlyShort
     */
    DBR_GR_SHORT(22),

    /**
     * Type representing a single-precision floating-point value with alarm
     * information, engineering units, display precision, and alarm and display
     * limits.
     * 
     * @see ChannelAccessGraphicsOnlyFloat
     */
    DBR_GR_FLOAT(23),

    /**
     * Type representing an enum value with alarm information and labels.
     * Internally, an enum is simply a short, but it is interpreted as a set of
     * states to choose from.
     * 
     * @see ChannelAccessGraphicsEnum
     */
    DBR_GR_ENUM(24),

    /**
     * Type representing a simple byte value with alarm information, engineering
     * units, and alarm and display limits. For historical reasons, this type is
     * called <code>DBR_GR_CHAR</code>, however it represents an 8-bit integer (
     * <code>byte</code> in Java). Channel Access specifies that this type is
     * unsigned, but due to limitations of the Java platform, we us the signed
     * version.
     * 
     * @see ChannelAccessGraphicsOnlyChar
     */
    DBR_GR_CHAR(25),

    /**
     * Type representing an integer value with alarm information, engineering
     * units, and alarm and display limits. For historical reasons, this type is
     * called <code>DBR_GR_LONG</code>, however it represents a 32-bit integer (
     * <code>int</code> in Java).
     * 
     * @see ChannelAccessGraphicsOnlyLong
     */
    DBR_GR_LONG(26),

    /**
     * Type representing a double-precision floating-point value with alarm
     * information, engineering units, display precision, and alarm and display
     * limits.
     * 
     * @see ChannelAccessGraphicsOnlyDouble
     */
    DBR_GR_DOUBLE(27),

    /**
     * Type representing a string value with alarm information. Unlike other
     * data-types, strings do not have any additional graphics or controls
     * information, so the representation of this type is identical to the one
     * used by {@link #DBR_STS_STRING}.
     * 
     * @see ChannelAccessAlarmOnlyString
     */
    DBR_CTRL_STRING(28),

    /**
     * Type representing a short integer value with alarm information,
     * engineering units, and alarm, control, and display limits.
     * 
     * @see ChannelAccessControlsShort
     */
    DBR_CTRL_SHORT(29),

    /**
     * Type representing a single-precision floating-point value with alarm
     * information, engineering units, display precision, and alarm, control,
     * and display limits.
     * 
     * @see ChannelAccessControlsFloat
     */
    DBR_CTRL_FLOAT(30),

    /**
     * Type representing an enum value with alarm information and labels. Unlike
     * other data-types, enums do not have any addotional controls information,
     * so the representation of this type is identical to the one used by
     * {@link #DBR_GR_ENUM}. Internally, an enum is simply a short, but it is
     * interpreted as a set of states to choose from.
     * 
     * @see ChannelAccessGraphicsEnum
     */
    DBR_CTRL_ENUM(31),

    /**
     * Type representing a simple byte value with alarm information, engineering
     * units, and alarm, control, and display limits. For historical reasons,
     * this type is called <code>DBR_CTRL_CHAR</code>, however it represents an
     * 8-bit integer ( <code>byte</code> in Java). Channel Access specifies that
     * this type is unsigned, but due to limitations of the Java platform, we us
     * the signed version.
     * 
     * @see ChannelAccessControlsChar
     */
    DBR_CTRL_CHAR(32),

    /**
     * Type representing an integer value with alarm information, engineering
     * units, and alarm, control, and display limits. For historical reasons,
     * this type is called <code>DBR_CTRL_LONG</code>, however it represents a
     * 32-bit integer ( <code>int</code> in Java).
     * 
     * @see ChannelAccessControlsLong
     */
    DBR_CTRL_LONG(33),

    /**
     * Type representing a double-precision floating-point value with alarm
     * information, engineering units, display precision, and alarm, controls,
     * and display limits.
     * 
     * @see ChannelAccessControlsDouble
     */
    DBR_CTRL_DOUBLE(34),

    /**
     * Type used to configure the transient alarm setting. This type is only
     * used in put operations in order to configure whether transient alarms for
     * a channel should have to be acknowledged explicitly or should be
     * acknowledged automatically, when they disappear.
     * 
     * @see ChannelAccessConfigureAcknowledgeTransientAlarms
     */
    DBR_PUT_ACKT(35),

    /**
     * Type used to acknowledge alarms. This type is only used in put operations
     * in order to acknowledge an alarm for a channel.
     * 
     * @see ChannelAccessAcknowledgeAlarm
     */
    DBR_PUT_ACKS(36),

    /**
     * Type used to retrieve information about acknowledged and unacknowledged
     * alarms. This type is only used in get or monitor operations and contains
     * information about the current alarm state of a channel, whether alarms
     * are acknowledged or unacknowledged, and whether transient alarms have to
     * be acknowledged explicitly.
     * 
     * @see ChannelAccessAlarmAcknowledgementStatus
     */
    DBR_STSACK_STRING(37),

    /**
     * Types used to retrieve information about the underlying record type of a
     * channel. This type is only used in get operations in order to get the
     * record type of a channel. Depending on the kind of Channel Access server,
     * the record type might not be well-defined.
     * 
     * @see ChannelAccessClassName
     */
    DBR_CLASS_NAME(38);

    private static final ChannelAccessValueType[] typeByCode;

    static {
        ChannelAccessValueType[] allValues = ChannelAccessValueType.values();
        typeByCode = new ChannelAccessValueType[allValues.length];
        for (ChannelAccessValueType valueType : allValues) {
            typeByCode[valueType.toTypeCode()] = valueType;
        }
    }

    private short typeCode;
    private boolean simpleType;
    private boolean alarmType;
    private boolean timeType;
    private boolean graphicsType;
    private boolean controlsType;
    private boolean specialType;

    private ChannelAccessValueType(int typeCode) {
        // We cast the type number here so that the definition of the various
        // types is kept simple. We can do this because we know all passed
        // values are within the allowed range.
        this.typeCode = (short) typeCode;
        this.simpleType = (this.typeCode < 7);
        this.alarmType = (this.typeCode >= 7) && (this.typeCode < 14);
        this.timeType = (this.typeCode >= 14) && (this.typeCode < 21);
        this.graphicsType = (this.typeCode >= 21) && (this.typeCode < 28);
        this.controlsType = (this.typeCode >= 28) && (this.typeCode < 35);
        this.specialType = (this.typeCode >= 35);
    }

    /**
     * Returns the type number. The type number is a numeric identifier that is
     * used to identify a data type in the over-the-wire protocol.
     * 
     * @return numeric identifier for this type.
     * @see #forTypeNumber(short)
     */
    public short toTypeCode() {
        return typeCode;
    }

    /**
     * Returns whether this type represents a simple type. A simple type is a
     * type which only stores a value and does not provide additional
     * information (e.g. the alarm status). The {@link #DBR_PUT_ACKT},
     * {@link #DBR_PUT_ACKS}, {@link #DBR_STSACK_STRING}, and
     * {@link #DBR_CLASS_NAME} types are not considers simple types because they
     * do not transport plain values but have special meanings. Therefore, this
     * method returns <code>false</code> for these types.
     * 
     * @return <code>true</code> if this type represents a simple type,
     *         <code>false</code> otherwise.
     */
    public boolean isSimpleType() {
        return simpleType;
    }

    /**
     * Returns whether this type represents a type with alarm information (but
     * no additional information). This is <code>true</code> for the
     * <code>DBR_STS_*</code> types. This method returns <code>false</code> for
     * the {@link #DBR_STSACK_STRING} type, because it is considered a special,
     * even though it transports alarm information.
     * 
     * @return <code>true</code> if this type represent a (pure) type with alarm
     *         information, <code>false</code> otherwise.
     * @see #includesAlarm()
     */
    public boolean isAlarmType() {
        return alarmType;
    }

    /**
     * Returns whether this type represents a type with a time-stamp. This is
     * <code>true</code> for the <code>DBR_TIME_*</code> types.
     * 
     * @return <code>true</code> if this type represents a type with alarm
     *         information, <code>false</code> otherwise.
     */
    public boolean isTimeType() {
        return timeType;
    }

    /**
     * Returns <code>true</code> if this type represents a type with alarm and
     * display information. This is <code>true</code> for the
     * <code>DBR_GR_*</code> types, including {@link #DBR_GR_STRING}, even
     * though it does not provide control limits.
     * 
     * @return <code>true</code> if this type represents a type with alarm and
     *         display information, <code>false</code> otherwise.
     * @see #includesGraphics()
     */
    public boolean isGraphicsType() {
        return graphicsType;
    }

    /**
     * Returns <code>true</code> if this type represents a type with alarm,
     * display, and control limits information. This is <code>true</code> for
     * the <code>DBR_CTRL*</code> types, including {@link #DBR_CTRL_STRING},
     * even though it does not provide control limits.
     * 
     * @return <code>true</code> if this type represents a type with alarm,
     *         display, and control limits information, <code>false</code>
     *         otherwise.
     */
    public boolean isControlsType() {
        return controlsType;
    }

    /**
     * Returns <code>true</code> if this type is one of the special types,
     * <code>false</code> otherwise. {@link #DBR_PUT_ACKT},
     * {@link #DBR_PUT_ACKS}, {@link #DBR_STSACK_STRING}, and
     * {@link #DBR_CLASS_NAME} are considered special types because they cannot
     * be converted between simple, alarm, time, graphics, and controls types.
     * 
     * @return <code>true</code> if this type is a @link #DBR_PUT_ACKT},
     *         {@link #DBR_PUT_ACKS}, {@link #DBR_STSACK_STRING}, or
     *         {@link #DBR_CLASS_NAME}, <code>false</code> otherwise.
     */
    public boolean isSpecialType() {
        return specialType;
    }

    /**
     * Returns <code>true</code> if this type can be used in a get operation
     * (transferring data from a server to a client), <code>false</code>
     * otherwise. The regular types ({@link #isSpecialType()}
     * <code>== false</code>), {@link #DBR_STSACK_STRING}, and
     * {@link #DBR_CLASS_NAME} are gettable. Only {@link #DBR_PUT_ACKT} and
     * {@link #DBR_PUT_ACKS} are not gettable.
     * 
     * @return <code>true</code> if this type is gettable, <code>false</code>
     *         otherwise.
     */
    public boolean isGettable() {
        return typeCode != 35 && typeCode != 36;
    }

    /**
     * Returns <code>true</code> if this type can be used in a put
     * operation(transferring data from a client to a server),
     * <code>false</code> otherwise. The simple types ({@link #isSimpleType()}
     * <code>== true</code>), {@link #DBR_PUT_ACKT}, and {@link #DBR_PUT_ACKS}
     * are puttable.
     * 
     * @return <code>true</code> if this type is puttable, <code>false</code>
     *         otherwise.
     */
    public boolean isPuttable() {
        return isSimpleType() || typeCode == 35 || typeCode == 36;
    }

    /**
     * Returns <code>true</code> if this type includes alarm information. This
     * is <code>true</code> for the <code>DBR_STS_*</code>,
     * <code>DBR_TIME_*</code>, <code>DBR_GR_*</code>, and
     * <code>DBR_CTRL_*</code> types. This method returns <code>false</code> for
     * the {@link #DBR_STSACK_STRING} type, because it is considered a simple
     * type, even though it transports alarm information.
     * 
     * @return <code>true</code> if this type provides alarm information,
     *         <code>false</code> otherwise.
     * @see #isAlarmType()
     */
    public boolean includesAlarm() {
        return alarmType || timeType || graphicsType || controlsType;
    }

    /**
     * Returns <code>true</code> if this type includes display information. This
     * is <code>true</code> for the <code>DBR_GR_*</code> and
     * <code>DBR_CTRL_*</code> types, including the {@link #DBR_GR_STRING} and
     * {@link #DBR_CTRL_STRING} types, even though they do not provide display
     * information.
     * 
     * @return <code>true</code> if this type provides display information,
     *         <code>false</code> otherwise.
     * @see #isGraphicsType()
     */
    public boolean includesGraphics() {
        return graphicsType || controlsType;
    }

    /**
     * Returns the simple type that corresponds to this type (has the same value
     * type). If this type is already a simple type, the same type is returned.
     * 
     * @return simple type corresponding to this type.
     * @see #isSimpleType()
     */
    public ChannelAccessValueType toSimpleType() {
        if (typeCode >= 35) {
            throw new IllegalArgumentException(
                    "Type "
                            + toString()
                            + " is a special type and cannot be converted to a \"simple\" type.");
        } else {
            return typeByCode[typeCode % 7];
        }
    }

    /**
     * Returns the alarm type that corresponds to this type (has the same value
     * type). If this type is already an alarm type, the same type is returned.
     * Calling this method for the {@link #DBR_PUT_ACKT}, {@link #DBR_PUT_ACKS},
     * {@link #DBR_STSACK_STRING}, or {@link #DBR_CLASS_NAME} type results in an
     * {@link IllegalArgumentException} being thrown, because there are no
     * corresponding alarm types for these types.
     * 
     * @return alarm type corresponding to this type.
     * @throws IllegalArgumentException
     *             if there is no alarm type corresponding to this type.
     * @see #isAlarmType()
     */
    public ChannelAccessValueType toAlarmType() {
        if (typeCode >= 35) {
            throw new IllegalArgumentException("Type " + toString()
                    + " does not have a corresponding DBR_STS_* type.");
        } else {
            return typeByCode[typeCode % 7 + 7];
        }
    }

    /**
     * Returns the time type that corresponds to this type (has the same value
     * type). If this type is already a time type, the same type is returned.
     * Calling this method for the {@link #DBR_PUT_ACKT}, {@link #DBR_PUT_ACKS},
     * {@link #DBR_STSACK_STRING}, or {@link #DBR_CLASS_NAME} type results in an
     * {@link IllegalArgumentException} being thrown, because there are no
     * corresponding time types for these types.
     * 
     * @return time type corresponding to this type.
     * @throws IllegalArgumentException
     *             if there is no time type corresponding to this type.
     * @see #isTimeType()
     */
    public ChannelAccessValueType toTimeType() {
        if (typeCode >= 35) {
            throw new IllegalArgumentException("Type " + toString()
                    + " does not have a corresponding DBR_TIME_* type.");
        } else {
            return typeByCode[typeCode % 7 + 14];
        }
    }

    /**
     * Returns the graphics type that corresponds to this type (has the same
     * value type). If this type is already a graphics type, the same type is
     * returned. Calling this method for the {@link #DBR_PUT_ACKT},
     * {@link #DBR_PUT_ACKS}, {@link #DBR_STSACK_STRING}, or
     * {@link #DBR_CLASS_NAME} type results in an
     * {@link IllegalArgumentException} being thrown, because there are no
     * corresponding graphics types for these types.
     * 
     * @return graphics type corresponding to this type.
     * @throws IllegalArgumentException
     *             if there is no graphics type corresponding to this type.
     * @see #isTimeType()
     */
    public ChannelAccessValueType toGraphicsType() {
        if (typeCode >= 35) {
            throw new IllegalArgumentException("Type " + toString()
                    + " does not have a corresponding DBR_GR_* type.");
        } else {
            return typeByCode[typeCode % 7 + 21];
        }
    }

    /**
     * Returns the controls type that corresponds to this type (has the same
     * value type). If this type is already a controls type, the same type is
     * returned. Calling this method for the {@link #DBR_PUT_ACKT},
     * {@link #DBR_PUT_ACKS}, {@link #DBR_STSACK_STRING}, or
     * {@link #DBR_CLASS_NAME} type results in an
     * {@link IllegalArgumentException} being thrown, because there are no
     * corresponding controls types for these types.
     * 
     * @return controls type corresponding to this type.
     * @throws IllegalArgumentException
     *             if there is no controls type corresponding to this type.
     * @see #isTimeType()
     */
    public ChannelAccessValueType toControlsType() {
        if (typeCode >= 35) {
            throw new IllegalArgumentException("Type " + toString()
                    + " does not have a corresponding DBR_CTRL_* type.");
        } else {
            return typeByCode[typeCode % 7 + 28];
        }
    }

    /**
     * Returns the Channel Access data-type that is identified by the specified
     * type number. Each data-type has a unique type number that is used to
     * identify this type in the over-the-wire protocol.
     * 
     * @param typeNumber
     *            type number identifying the requested type.
     * @return type identified by the specified type number (never
     *         <code>null</code>).
     * @throws IllegalArgumentException
     *             if the specified number does not identify a valid Channel
     *             Access data-type.
     * @see #toTypeCode()
     */
    public static ChannelAccessValueType forTypeNumber(short typeNumber) {
        if (typeNumber < 0 || typeNumber > 38) {
            throw new IllegalArgumentException(
                    "Type number for a Channel Access type must be between 0 and 38.");
        }
        return typeByCode[typeNumber];
    }

}
