/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.ShortBuffer;
import java.nio.charset.Charset;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeStamp;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Base class for Channel Access short type. This class contains the various
 * specializations of the short type as inlined classes.
 * 
 * @param <ImplementationType>
 *            type of the actual implementation.
 * 
 * @author Sebastian Marsching
 */
public abstract class ChannelAccessShortBase<ImplementationType extends ChannelAccessShortBase<ImplementationType>>
        extends ChannelAccessValueBase<Short, ImplementationType> implements
        ChannelAccessShort {

    /**
     * Implementation of DBR_SHORT.
     */
    public static final class ChannelAccessShortImpl extends
            ChannelAccessShortBase<ChannelAccessShortImpl> implements
            ChannelAccessSimpleOnlyShort {

        /**
         * Creates a DBR_SHORT.
         * 
         * @param value
         *            short value.
         */
        public ChannelAccessShortImpl(short[] value) {
            super(value);
        }

        /**
         * Creates a read-only DBR_SHORT. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessShortImpl(ChannelAccessShortImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessShortImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_SHORT;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessSimpleOnlyShort asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessShortImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessSimpleOnlyShort clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessSimpleOnlyShort) super.clone();
        }

        /**
         * Creates a DBR_SHORT from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_SHORT instance created from the serialized data read from
         *         <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessShortImpl deserialize(ByteSource byteSource,
                int count) {
            ChannelAccessShortImpl value = new ChannelAccessShortImpl();
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for short with alarms.
     */
    public static abstract class ChannelAccessAlarmShortBase<ImplementationType extends ChannelAccessAlarmShortBase<ImplementationType>>
            extends ChannelAccessShortBase<ImplementationType> implements
            ChannelAccessAlarmShort {

        /**
         * Alarm status.
         */
        protected ChannelAccessAlarmStatus alarmStatus;

        /**
         * Alarm severity.
         */
        protected ChannelAccessAlarmSeverity alarmSeverity;

        /**
         * Base constructor.
         * 
         * @param value
         *            short value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        private ChannelAccessAlarmShortBase(short[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value);
            this.alarmStatus = alarmStatus;
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmShortBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmShortBase() {
        }

        @Override
        public ChannelAccessAlarmStatus getAlarmStatus() {
            if (delegate != null) {
                return delegate.getAlarmStatus();
            }
            return this.alarmStatus;
        }

        @Override
        public void setAlarmStatus(ChannelAccessAlarmStatus alarmStatus) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmStatus != null);
            this.alarmStatus = alarmStatus;
        }

        @Override
        public ChannelAccessAlarmSeverity getAlarmSeverity() {
            if (delegate != null) {
                return delegate.getAlarmSeverity();
            }
            return this.alarmSeverity;
        }

        @Override
        public void setAlarmSeverity(ChannelAccessAlarmSeverity alarmSeverity) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmSeverity != null);
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Deserializes the alarm status from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeAlarms(ByteSource byteSource) {
            this.alarmStatus = ValueCodecUtils
                    .deserializeAlarmStatus(byteSource);
            this.alarmSeverity = ValueCodecUtils
                    .deserializeAlarmSeverity(byteSource);
        }

        /**
         * Serializes the alarm status to the byte sink.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeAlarms(ByteSink byteSink) {
            ValueCodecUtils.serializeAlarms(byteSink, alarmStatus,
                    alarmSeverity);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessAlarmShortBase<ImplementationType> other = (ChannelAccessAlarmShortBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.alarmStatus, other.alarmStatus)
                    .append(this.alarmSeverity, other.alarmSeverity).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.alarmStatus).append(this.alarmSeverity)
                    .toHashCode();
        }

        @Override
        public ChannelAccessAlarmShort clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return (ChannelAccessAlarmShort) super.clone();
        }

    }

    /**
     * Implementation of DBR_STS_SHORT.
     */
    public static final class ChannelAccessAlarmShortImpl extends
            ChannelAccessAlarmShortBase<ChannelAccessAlarmShortImpl> implements
            ChannelAccessAlarmOnlyShort {

        /**
         * Creates a DBR_STS_SHORT.
         * 
         * @param value
         *            short value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        public ChannelAccessAlarmShortImpl(short[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value, alarmStatus, alarmSeverity);
        }

        /**
         * Creates a read-only DBR_STS_SHORT. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmShortImpl(ChannelAccessAlarmShortImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmShortImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_STS_SHORT;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessAlarmOnlyShort asReadOnlyValue() {
            if (delegate != null) {
                return this;
            } else {
                return new ChannelAccessAlarmShortImpl(this);
            }
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessAlarmOnlyShort clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessAlarmOnlyShort) super.clone();
        }

        /**
         * Creates a DBR_STS_SHORT from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_STS_SHORT instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessAlarmShortImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessAlarmShortImpl value = new ChannelAccessAlarmShortImpl();
            value.deserializeAlarms(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_TIME_SHORT.
     */
    public static final class ChannelAccessTimeShortImpl extends
            ChannelAccessAlarmShortBase<ChannelAccessTimeShortImpl> implements
            ChannelAccessTimeShort {

        /**
         * Seconds part of time-stamp.
         */
        protected int timeStampSeconds;

        /**
         * Nanoseconds part of time-stamp.
         */
        protected int timeStampNanoseconds;

        /**
         * Creates a DBR_TIME_SHORT.
         * 
         * @param value
         *            short value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param timeStampSeconds
         *            seconds part of time-stamp.
         * @param timeStampNanoSeconds
         *            nanoseconds part of time-stamp.
         */
        public ChannelAccessTimeShortImpl(short[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
                int timeStampNanoSeconds) {
            super(value, alarmStatus, alarmSeverity);
            this.timeStampSeconds = timeStampSeconds;
            if (timeStampNanoSeconds < 0 || timeStampNanoSeconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of timestamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoSeconds;
        }

        /**
         * Creates a read-only DBR_TIME_SHORT. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessTimeShortImpl(ChannelAccessTimeShortImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessTimeShortImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_TIME_SHORT;
        }

        @Override
        public ChannelAccessTimeStamp getTimeStamp() {
            if (delegate != null) {
                return delegate.getTimeStamp();
            }
            return ChannelAccessValueFactory.createTimeStamp(timeStampSeconds,
                    timeStampNanoseconds);
        }

        @Override
        public int getTimeSeconds() {
            if (delegate != null) {
                return delegate.getTimeSeconds();
            }
            return timeStampSeconds;
        }

        @Override
        public int getTimeNanoseconds() {
            if (delegate != null) {
                return delegate.getTimeNanoseconds();
            }
            return timeStampNanoseconds;
        }

        @Override
        public void setTimeStamp(ChannelAccessTimeStamp timeStamp) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStamp.getSeconds();
            this.timeStampNanoseconds = timeStamp.getNanoseconds();
        }

        @Override
        public void setTimeSeconds(int timeStampSeconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStampSeconds;
        }

        @Override
        public void setTimeNanoseconds(int timeStampNanoseconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            if (timeStampNanoseconds < 0 || timeStampNanoseconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of time-stamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoseconds;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            ValueCodecUtils.serializeTime(byteSink, timeStampSeconds,
                    timeStampNanoseconds);
            // Insert two extra padding bytes.
            byteSink.putShort((short) 0);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessTimeShort asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessTimeShortImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessTimeShortImpl other = (ChannelAccessTimeShortImpl) obj;
            return new EqualsBuilder()
                    .append(this.timeStampSeconds, other.timeStampSeconds)
                    .append(this.timeStampNanoseconds,
                            other.timeStampNanoseconds).isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.timeStampSeconds)
                    .append(this.timeStampNanoseconds).toHashCode();
        }

        @Override
        public ChannelAccessTimeShort clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessTimeShort) super.clone();
        }

        /**
         * Creates a DBR_TIME_SHORT from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_TIME_SHORT instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessTimeShortImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessTimeShortImpl value = new ChannelAccessTimeShortImpl();
            value.deserializeAlarms(byteSource);
            // Deserialize time-stamp.
            value.timeStampSeconds = ValueCodecUtils
                    .deserializeTimeSeconds(byteSource);
            value.timeStampNanoseconds = ValueCodecUtils
                    .deserializeTimeNanoseconds(byteSource);
            // Skip two extra padding bytes.
            byteSource.skip(2);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for short with display information.
     */
    public static abstract class ChannelAccessGraphicsShortBase<ImplementationType extends ChannelAccessGraphicsShortBase<ImplementationType>>
            extends ChannelAccessAlarmShortBase<ImplementationType> implements
            ChannelAccessGraphicsShort {

        private static final int ENGINEERING_UNITS_STRING_SIZE = 8;

        /**
         * Engineering units.
         */
        protected String engineeringUnits;

        /**
         * Raw representation of engineering units.
         */
        protected byte[] rawEngineeringUnits;

        /**
         * Upper display limit.
         */
        protected short upperDisplayLimit;

        /**
         * Lower display limit.
         */
        protected short lowerDisplayLimit;

        /**
         * Upper alarm limit.
         */
        protected short upperAlarmLimit;

        /**
         * Upper warning limit.
         */
        protected short upperWarningLimit;

        /**
         * Lower warning limit.
         */
        protected short lowerWarningLimit;

        /**
         * Lower alarm limit.
         */
        protected short lowerAlarmLimit;

        /**
         * Charset used for encoding/decoding strings.
         */
        protected Charset charset;

        /**
         * Base constructor.
         * 
         * @param value
         *            short value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsShortBase(short[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity,
                String engineeringUnits, short upperDisplayLimit,
                short lowerDisplayLimit, short upperAlarmLimit,
                short upperWarningLimit, short lowerWarningLimit,
                short lowerAlarmLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity);
            this.engineeringUnits = engineeringUnits;
            assert (charset != null);
            this.charset = charset;
            this.rawEngineeringUnits = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(
                            this.engineeringUnits,
                            ENGINEERING_UNITS_STRING_SIZE, charset);
            this.upperDisplayLimit = upperDisplayLimit;
            this.lowerDisplayLimit = lowerDisplayLimit;
            this.upperAlarmLimit = upperAlarmLimit;
            this.upperWarningLimit = upperWarningLimit;
            this.lowerWarningLimit = lowerWarningLimit;
            this.lowerAlarmLimit = lowerAlarmLimit;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsShortBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsShortBase(Charset charset) {
            this.charset = charset;
        }

        @Override
        public String getUnits() {
            if (delegate != null) {
                return delegate.getUnits();
            }
            return engineeringUnits;
        }

        @Override
        public byte[] getRawUnits() {
            if (delegate != null) {
                return delegate.getRawUnits();
            }
            return rawEngineeringUnits.clone();
        }

        @Override
        public void setUnits(String units) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.engineeringUnits = units;
            // Update raw representation.
            this.rawEngineeringUnits = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(
                            this.engineeringUnits,
                            ENGINEERING_UNITS_STRING_SIZE, charset);
        }

        @Override
        public void setRawUnits(byte[] rawUnits) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            // We use a new array for storing the raw units. This way, we do
            // not have to copy the array in the clone() method and we do not
            // have to fill the array with null bytes explicitly.
            this.rawEngineeringUnits = new byte[ENGINEERING_UNITS_STRING_SIZE];
            // Copy first seven bytes.
            int copyLength = Math.min(rawUnits.length,
                    ENGINEERING_UNITS_STRING_SIZE - 1);
            System.arraycopy(rawUnits, 0, this.rawEngineeringUnits, 0,
                    copyLength);
            // Update string.
            this.engineeringUnits = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(this.rawEngineeringUnits,
                            charset);
        }

        @Override
        public Short getGenericUpperDisplayLimit() {
            return getUpperDisplayLimit();
        }

        @Override
        public Short getGenericLowerDisplayLimit() {
            return getLowerDisplayLimit();
        }

        @Override
        public Short getGenericUpperAlarmLimit() {
            return getUpperAlarmLimit();
        }

        @Override
        public Short getGenericUpperWarningLimit() {
            return getUpperWarningLimit();
        }

        @Override
        public Short getGenericLowerWarningLimit() {
            return getLowerWarningLimit();
        }

        @Override
        public Short getGenericLowerAlarmLimit() {
            return getLowerAlarmLimit();
        }

        @Override
        public Charset getCharset() {
            if (delegate != null) {
                return delegate.getCharset();
            }
            return charset;
        }

        @Override
        public short getUpperDisplayLimit() {
            if (delegate != null) {
                return delegate.getUpperDisplayLimit();
            }
            return upperDisplayLimit;
        }

        @Override
        public void setUpperDisplayLimit(short upperDisplayLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperDisplayLimit = upperDisplayLimit;
        }

        @Override
        public short getLowerDisplayLimit() {
            if (delegate != null) {
                return delegate.getLowerDisplayLimit();
            }
            return lowerDisplayLimit;
        }

        @Override
        public void setLowerDisplayLimit(short lowerDisplayLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerDisplayLimit = lowerDisplayLimit;
        }

        @Override
        public short getUpperAlarmLimit() {
            if (delegate != null) {
                return delegate.getUpperAlarmLimit();
            }
            return upperAlarmLimit;
        }

        @Override
        public void setUpperAlarmLimit(short upperAlarmLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperAlarmLimit = upperAlarmLimit;
        }

        @Override
        public short getUpperWarningLimit() {
            if (delegate != null) {
                return delegate.getUpperWarningLimit();
            }
            return upperWarningLimit;
        }

        @Override
        public void setUpperWarningLimit(short upperWarningLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperWarningLimit = upperWarningLimit;
        }

        @Override
        public short getLowerWarningLimit() {
            if (delegate != null) {
                return delegate.getLowerWarningLimit();
            }
            return lowerWarningLimit;
        }

        @Override
        public void setLowerWarningLimit(short lowerWarningLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerWarningLimit = lowerWarningLimit;
        }

        @Override
        public short getLowerAlarmLimit() {
            if (delegate != null) {
                return delegate.getLowerAlarmLimit();
            }
            return lowerAlarmLimit;
        }

        @Override
        public void setLowerAlarmLimit(short lowerAlarmLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerAlarmLimit = lowerAlarmLimit;
        }

        /**
         * Deserializes the display information from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeGraphics(ByteSource byteSource) {
            this.rawEngineeringUnits = byteSource
                    .getByteArray(ENGINEERING_UNITS_STRING_SIZE);
            // Ensure that engineering units are null terminated.
            this.rawEngineeringUnits[ENGINEERING_UNITS_STRING_SIZE - 1] = 0;
            this.engineeringUnits = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(this.rawEngineeringUnits,
                            charset);
            this.upperDisplayLimit = byteSource.getShort();
            this.lowerDisplayLimit = byteSource.getShort();
            this.upperAlarmLimit = byteSource.getShort();
            this.upperWarningLimit = byteSource.getShort();
            this.lowerWarningLimit = byteSource.getShort();
            this.lowerAlarmLimit = byteSource.getShort();
        }

        /**
         * Serializes the display information.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeGraphics(ByteSink byteSink) {
            byteSink.putByteArray(rawEngineeringUnits);
            byteSink.putShort(upperDisplayLimit);
            byteSink.putShort(lowerDisplayLimit);
            byteSink.putShort(upperAlarmLimit);
            byteSink.putShort(upperWarningLimit);
            byteSink.putShort(lowerWarningLimit);
            byteSink.putShort(lowerAlarmLimit);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessGraphicsShortBase<ImplementationType> other = (ChannelAccessGraphicsShortBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.engineeringUnits, other.engineeringUnits)
                    .append(this.rawEngineeringUnits, other.rawEngineeringUnits)
                    .append(this.upperDisplayLimit, other.upperDisplayLimit)
                    .append(this.lowerDisplayLimit, other.lowerDisplayLimit)
                    .append(this.upperAlarmLimit, other.upperAlarmLimit)
                    .append(this.upperWarningLimit, other.upperWarningLimit)
                    .append(this.lowerWarningLimit, other.lowerWarningLimit)
                    .append(this.lowerAlarmLimit, other.lowerAlarmLimit)
                    .append(this.charset, other.charset).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.engineeringUnits)
                    .append(this.rawEngineeringUnits)
                    .append(this.upperDisplayLimit)
                    .append(this.lowerDisplayLimit)
                    .append(this.upperAlarmLimit)
                    .append(this.upperWarningLimit)
                    .append(this.lowerWarningLimit)
                    .append(this.lowerAlarmLimit).append(this.charset)
                    .toHashCode();
        }

        @Override
        public ChannelAccessGraphicsShort clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            // We do not have to do anything special because we never modify
            // the rawEngineeringUnits array.
            return (ChannelAccessGraphicsShort) super.clone();
        }

    }

    /**
     * Implementation of DBR_GR_SHORT.
     */
    public static final class ChannelAccessGraphicsShortImpl extends
            ChannelAccessGraphicsShortBase<ChannelAccessGraphicsShortImpl>
            implements ChannelAccessGraphicsOnlyShort {

        /**
         * Creates a DBR_GR_SHORT.
         * 
         * @param value
         *            short value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessGraphicsShortImpl(short[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity,
                String engineeringUnits, short upperDisplayLimit,
                short lowerDisplayLimit, short upperAlarmLimit,
                short upperWarningLimit, short lowerWarningLimit,
                short lowerAlarmLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity, engineeringUnits,
                    upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                    upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                    charset);
        }

        /**
         * Creates a read-only DBR_GR_SHORT. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsShortImpl(
                ChannelAccessGraphicsShortImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsShortImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_GR_SHORT;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeGraphics(byteSink);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessGraphicsOnlyShort asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessGraphicsShortImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessGraphicsOnlyShort clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessGraphicsOnlyShort) super.clone();
        }

        /**
         * Creates a DBR_GR_SHORT from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_GR_SHORT instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessGraphicsShortImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessGraphicsShortImpl value = new ChannelAccessGraphicsShortImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            value.deserializeGraphics(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_CTRL_SHORT.
     */
    public static final class ChannelAccessControlsShortImpl extends
            ChannelAccessGraphicsShortBase<ChannelAccessControlsShortImpl>
            implements ChannelAccessControlsShort {

        /**
         * Upper control limit.
         */
        protected short upperControlLimit;

        /**
         * Lower control limit.
         */
        protected short lowerControlLimit;

        /**
         * Creates a DBR_CTRL_SHORT.
         * 
         * @param value
         *            short value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param upperControlLimit
         *            upper control limit.
         * @param lowerControlLimit
         *            lower control limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessControlsShortImpl(short[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity,
                String engineeringUnits, short upperDisplayLimit,
                short lowerDisplayLimit, short upperAlarmLimit,
                short upperWarningLimit, short lowerWarningLimit,
                short lowerAlarmLimit, short upperControlLimit,
                short lowerControlLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity, engineeringUnits,
                    upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                    upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                    charset);
            this.upperControlLimit = upperControlLimit;
            this.lowerControlLimit = lowerControlLimit;
        }

        /**
         * Creates a read-only DBR_GR_SHORT. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessControlsShortImpl(
                ChannelAccessControlsShortImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessControlsShortImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_CTRL_SHORT;
        }

        @Override
        public Short getGenericUpperControlLimit() {
            return getUpperControlLimit();
        }

        @Override
        public Short getGenericLowerControlLimit() {
            return getLowerControlLimit();
        }

        @Override
        public short getUpperControlLimit() {
            if (delegate != null) {
                return delegate.getUpperControlLimit();
            }
            return upperControlLimit;
        }

        @Override
        public void setUpperControlLimit(short upperControlLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperControlLimit = upperControlLimit;
        }

        @Override
        public short getLowerControlLimit() {
            if (delegate != null) {
                return delegate.getLowerControlLimit();
            }
            return lowerControlLimit;
        }

        @Override
        public void setLowerControlLimit(short lowerControlLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerControlLimit = lowerControlLimit;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeGraphics(byteSink);
            byteSink.putShort(upperControlLimit);
            byteSink.putShort(lowerControlLimit);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessControlsShort asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessControlsShortImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessControlsShortImpl other = (ChannelAccessControlsShortImpl) obj;
            return new EqualsBuilder()
                    .append(this.upperControlLimit, other.upperControlLimit)
                    .append(this.lowerControlLimit, other.lowerControlLimit)
                    .isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.upperControlLimit)
                    .append(this.lowerControlLimit).toHashCode();
        }

        @Override
        public ChannelAccessControlsShort clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessControlsShort) super.clone();
        }

        /**
         * Creates a DBR_CTRL_SHORT from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_CTRL_SHORT instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessControlsShortImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessControlsShortImpl value = new ChannelAccessControlsShortImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            value.deserializeGraphics(byteSource);
            value.upperControlLimit = byteSource.getShort();
            value.lowerControlLimit = byteSource.getShort();
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    protected short[] value;

    /**
     * Base constructor.
     * 
     * @param value
     *            short value.
     */
    private ChannelAccessShortBase(short[] value) {
        assert (value != null);
        this.value = value;
    }

    /**
     * Base constructor for read-only variant.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessShortBase(ImplementationType delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     */
    private ChannelAccessShortBase() {
    }

    @Override
    public int getValueSize() {
        if (delegate != null) {
            return delegate.getValueSize();
        }
        return value.length;
    }

    @Override
    public Short getGenericValueElement(int index) {
        if (delegate != null) {
            return delegate.getGenericValueElement(index);
        }
        return value[index];
    }

    @Override
    public ShortBuffer getValue() {
        if (delegate != null) {
            return delegate.getValue().asReadOnlyBuffer();
        }
        return ShortBuffer.wrap(value);
    }

    @Override
    public void setValue(short[] value) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        if (value == null) {
            throw new NullPointerException();
        }
        this.value = value;
    }

    /**
     * Deserializes the value elements from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @param count
     *            number of elements to deserialize.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    protected void deserializeValue(ByteSource byteSource, int count) {
        assert (count >= 0);
        if (count == 0) {
            byteSource.getShort();
            value = ArrayUtils.EMPTY_SHORT_ARRAY;
        } else {
            value = byteSource.getShortArray(count);
        }
    }

    /**
     * Serializes the value elements to the byte sink.
     * 
     * @param byteSink
     *            byte sink to write to.
     * @param count
     *            number of elements to be serialized.
     */
    protected void serializeValue(ByteSink byteSink, int count) {
        if (count == 0) {
            byteSink.putShort((short) 0);
        } else {
            byteSink.putShortArray(value, 0, count);
        }
    }

    @Override
    public boolean equals(Object obj) {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        if (!super.equals(obj)) {
            return false;
        }
        // Now, we can safely cast because the type has already been
        // checked.
        @SuppressWarnings("unchecked")
        ChannelAccessShortBase<ImplementationType> other = (ChannelAccessShortBase<ImplementationType>) obj;
        return new EqualsBuilder().append(this.value, other.value).isEquals();
    }

    @Override
    public int hashCode() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(value).toHashCode();
    }

    @Override
    public ChannelAccessShort clone() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        @SuppressWarnings("unchecked")
        ChannelAccessShortBase<ImplementationType> copy = (ChannelAccessShortBase<ImplementationType>) super
                .clone();
        copy.value = this.value.clone();
        return copy;
    }

}
