/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import static org.junit.Assert.*;

import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import org.junit.Test;

import com.aquenos.epics.jackie.common.io.ByteSource.AtomicGetOperation;

public class ByteBufferByteSourceTest {

    private static byte[] input1 = new byte[] { 1, 2, 3, 4, 5, 6 };
    private static byte[] input2 = new byte[] { 7 };
    private static byte[] input3 = new byte[] { 8, 9, 10, 11 };
    private static byte[] input4 = new byte[] { 12, 13, 14 };
    private static byte[] input5 = new byte[] { 15, 16 };

    private static ByteBufferByteSource sourceForByteArrays(
            byte[]... byteArrays) {
        ByteBufferByteSource source = new ByteBufferByteSource();
        for (byte[] byteArray : byteArrays) {
            source.appendBuffer(ByteBuffer.wrap(byteArray));
        }
        return source;
    }

    private static ByteBufferByteSource sourceForByteArraysLittleEndian(
            byte[]... byteArrays) {
        ByteBufferByteSource source = new ByteBufferByteSource();
        for (byte[] byteArray : byteArrays) {
            source.appendBuffer(ByteBuffer.wrap(byteArray).order(
                    ByteOrder.LITTLE_ENDIAN));
        }
        return source;
    }

    private static ByteBufferByteSource sourceForInputArrays1To5() {
        return sourceForByteArrays(input1, input2, input3, input4, input5);
    }

    private static ByteBufferByteSource sourceForInputArrays1To5LittleEndian() {
        return sourceForByteArraysLittleEndian(input1, input2, input3, input4,
                input5);
    }

    @Test
    public void readIncreasingBytes() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        assertEquals(16, source.remaining());
        for (int i = 0; i < 16; i++) {
            byte b = source.getByte();
            assertEquals(i + 1, b);
        }
        assertEquals(0, source.remaining());
    }

    @Test
    public void readIncreasingBytesAsArray() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        byte[] expectedArray = new byte[16];
        for (int i = 0; i < 16; i++) {
            expectedArray[i] = (byte) (i + 1);
        }
        byte[] readArray = source.getByteArray(16);
        assertArrayEquals(expectedArray, readArray);
    }

    @Test
    public void copyToDestinationArray() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        byte[] expectedArray = new byte[21];
        byte[] destinationArray = new byte[21];
        for (int i = 0; i < 16; i++) {
            expectedArray[i + 5] = (byte) (i + 1);
        }
        source.getByteArray(destinationArray, 5, 16);
        assertArrayEquals(expectedArray, destinationArray);
    }

    @Test
    public void readLongsSpanningBufferBorders() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        long l1 = source.getLong();
        long l2 = source.getLong();
        assertEquals(72623859790382856L, l1);
        assertEquals(651345242494996240L, l2);
    }

    @Test
    public void readLongsLittleEndian() {
        ByteBufferByteSource source = sourceForInputArrays1To5LittleEndian();
        long l1 = source.getLong();
        long l2 = source.getLong();
        assertEquals(578437695752307201L, l1);
        assertEquals(1157159078456920585L, l2);
    }

    @Test
    public void atomicReadWithError() {
        final ByteBufferByteSource source = sourceForInputArrays1To5();
        assertEquals(16, source.remaining());
        try {
            source.atomicGet(new AtomicGetOperation<Void>() {
                @Override
                public Void get() {
                    source.skip(8);
                    long l = source.getLong();
                    assertEquals(651345242494996240L, l);
                    throw new RuntimeException("expected exception");
                }
            });
        } catch (RuntimeException e) {
            assertEquals("expected exception", e.getMessage());
        }
        long l1 = source.getLong();
        assertEquals(8, source.remaining());
        long l2 = source.getLong();
        assertEquals(0, source.remaining());
        assertEquals(72623859790382856L, l1);
        assertEquals(651345242494996240L, l2);
    }

    @Test
    public void skipAndReadShorts() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        assertEquals(16, source.remaining());
        source.skip(5);
        assertEquals(11, source.remaining());
        short s = source.getShort();
        assertEquals(1543, s);
        assertEquals(9, source.remaining());
    }

    @Test(expected = BufferUnderflowException.class)
    public void readBeyondLimit() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        assertEquals(16, source.remaining());
        source.skip(15);
        // Should throw BufferUnderflowException
        source.getShort();
    }

    @Test(expected = IllegalArgumentException.class)
    public void negativeLength() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        // Should throw IllegalArgumentException
        source.getIntArray(-2);
    }

    @Test
    public void zeroLengthRead() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        assertEquals(16, source.remaining());
        source.getDoubleArray(0);
        assertEquals(16, source.remaining());
    }

    @Test(expected = IndexOutOfBoundsException.class)
    public void tooSmallDestinationArray() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        byte[] byteArray = new byte[5];
        source.getByteArray(byteArray, 2, 16);
    }

    @Test
    public void exceptionDoesNotChangePosition() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        assertEquals(16, source.remaining());
        source.skip(9);
        assertEquals(7, source.remaining());
        try {
            source.getIntArray(2);
        } catch (BufferUnderflowException e) {
            // We expect this exception and just want to check that the position
            // did not change.
        }
        assertEquals(7, source.remaining());
    }

    /**
     * Test that checks that the <code>remaining</code> methods returns the
     * correct number of bytes while an atomic read is in progress.
     */
    @Test
    public void atomicReadRemainingBytes() {
        final ByteBufferByteSource source = sourceForInputArrays1To5();
        assertEquals(16, source.remaining());
        source.atomicGet(new AtomicGetOperation<Void>() {
            @Override
            public Void get() {
                assertEquals(16, source.remaining());
                source.skip(8);
                assertEquals(8, source.remaining());
                source.getInt();
                assertEquals(4, source.remaining());
                return null;
            }
        });
        assertEquals(4, source.remaining());
    }

    /**
     * Test that the removing an appended buffer works correctly.
     */
    @Test
    public void removeLastBuffer() {
        ByteBufferByteSource source = sourceForInputArrays1To5();
        ByteBuffer buffer = source.removeLastBuffer();
        assertEquals(2, buffer.remaining());
        source.appendBuffer(buffer);
        source.skip(15);
        buffer = source.removeLastBuffer();
        assertEquals(1, buffer.remaining());
        source.appendBuffer(buffer);
        source.skip(1);
        buffer = source.removeLastBuffer();
        assertNull(buffer);
    }

    /**
     * Test that trying to remove a buffer while an atomic read operation is in
     * progress results in an exception.
     */
    @Test(expected = IllegalStateException.class)
    public void removeLastBufferAtomicReadFail() {
        final ByteBufferByteSource source = sourceForInputArrays1To5();
        source.atomicGet(new AtomicGetOperation<Void>() {
            @Override
            public Void get() {
                // Should throw
                source.removeLastBuffer();
                return null;
            }
        });
    }

}
